import inspect
import math
import random
import numbers
from numbers import Number
from typing import Sequence
import numpy as np
from PIL import Image
import cv2
from ..builder import PIPELINES


def imflip(img, direction='horizontal'):
    """Flip an image horizontally or vertically.

    Args:
        img (ndarray): Image to be flipped.
        direction (str): The flip direction, either "horizontal" or
            "vertical" or "diagonal".

    Returns:
        ndarray: The flipped image.
    """
    assert direction in ['horizontal', 'vertical', 'diagonal']
    if direction == 'horizontal':
        return np.flip(img, axis=1)
    elif direction == 'vertical':
        return np.flip(img, axis=0)
    else:
        return np.flip(img, axis=(0, 1))


def bbox_clip(bboxes, img_shape):
    """Clip bboxes to fit the image shape.

    Args:
        bboxes (ndarray): Shape (..., 4*k)
        img_shape (tuple[int]): (height, width) of the image.

    Returns:
        ndarray: Clipped bboxes.
    """
    assert bboxes.shape[-1] % 4 == 0
    cmin = np.empty(bboxes.shape[-1], dtype=bboxes.dtype)
    cmin[0::2] = img_shape[1] - 1
    cmin[1::2] = img_shape[0] - 1
    clipped_bboxes = np.maximum(np.minimum(bboxes, cmin), 0)
    return clipped_bboxes

def bbox_scaling(bboxes, scale, clip_shape=None):
    """Scaling bboxes w.r.t the box center.

    Args:
        bboxes (ndarray): Shape(..., 4).
        scale (float): Scaling factor.
        clip_shape (tuple[int], optional): If specified, bboxes that exceed the
            boundary will be clipped according to the given shape (h, w).

    Returns:
        ndarray: Scaled bboxes.
    """
    if float(scale) == 1.0:
        scaled_bboxes = bboxes.copy()
    else:
        w = bboxes[..., 2] - bboxes[..., 0] + 1
        h = bboxes[..., 3] - bboxes[..., 1] + 1
        dw = (w * (scale - 1)) * 0.5
        dh = (h * (scale - 1)) * 0.5
        scaled_bboxes = bboxes + np.stack((-dw, -dh, dw, dh), axis=-1)
    if clip_shape is not None:
        return bbox_clip(scaled_bboxes, clip_shape)
    else:
        return scaled_bboxes


def imcrop(img, bboxes, scale=1.0, pad_fill=None):
    """Crop image patches.

    3 steps: scale the bboxes -> clip bboxes -> crop and pad.

    Args:
        img (ndarray): Image to be cropped.
        bboxes (ndarray): Shape (k, 4) or (4, ), location of cropped bboxes.
        scale (float, optional): Scale ratio of bboxes, the default value
            1.0 means no padding.
        pad_fill (Number | list[Number]): Value to be filled for padding.
            Default: None, which means no padding.

    Returns:
        list[ndarray] | ndarray: The cropped image patches.
    """
    chn = 1 if img.ndim == 2 else img.shape[2]
    if pad_fill is not None:
        if isinstance(pad_fill, (int, float)):
            pad_fill = [pad_fill for _ in range(chn)]
        assert len(pad_fill) == chn

    _bboxes = bboxes[None, ...] if bboxes.ndim == 1 else bboxes
    scaled_bboxes = bbox_scaling(_bboxes, scale).astype(np.int32)
    clipped_bbox = bbox_clip(scaled_bboxes, img.shape)

    patches = []
    for i in range(clipped_bbox.shape[0]):
        x1, y1, x2, y2 = tuple(clipped_bbox[i, :])
        if pad_fill is None:
            patch = img[y1:y2 + 1, x1:x2 + 1, ...]
        else:
            _x1, _y1, _x2, _y2 = tuple(scaled_bboxes[i, :])
            if chn == 1:
                patch_shape = (_y2 - _y1 + 1, _x2 - _x1 + 1)
            else:
                patch_shape = (_y2 - _y1 + 1, _x2 - _x1 + 1, chn)
            patch = np.array(
                pad_fill, dtype=img.dtype) * np.ones(
                    patch_shape, dtype=img.dtype)
            x_start = 0 if _x1 >= 0 else -_x1
            y_start = 0 if _y1 >= 0 else -_y1
            w = x2 - x1 + 1
            h = y2 - y1 + 1
            patch[y_start:y_start + h, x_start:x_start + w,
                  ...] = img[y1:y1 + h, x1:x1 + w, ...]
        patches.append(patch)

    if bboxes.ndim == 1:
        return patches[0]
    else:
        return patches


def impad(img,
          *,
          shape=None,
          padding=None,
          pad_val=0,
          padding_mode='constant'):
    """Pad the given image to a certain shape or pad on all sides with
    specified padding mode and padding value.

    Args:
        img (ndarray): Image to be padded.
        shape (tuple[int]): Expected padding shape (h, w). Default: None.
        padding (int or tuple[int]): Padding on each border. If a single int is
            provided this is used to pad all borders. If tuple of length 2 is
            provided this is the padding on left/right and top/bottom
            respectively. If a tuple of length 4 is provided this is the
            padding for the left, top, right and bottom borders respectively.
            Default: None. Note that `shape` and `padding` can not be both
            set.
        pad_val (Number | Sequence[Number]): Values to be filled in padding
            areas when padding_mode is 'constant'. Default: 0.
        padding_mode (str): Type of padding. Should be: constant, edge,
            reflect or symmetric. Default: constant.
            - constant: pads with a constant value, this value is specified
              with pad_val.
            - edge: pads with the last value at the edge of the image.
            - reflect: pads with reflection of image without repeating the last
              value on the edge. For example, padding [1, 2, 3, 4] with 2
              elements on both sides in reflect mode will result in
              [3, 2, 1, 2, 3, 4, 3, 2].
            - symmetric: pads with reflection of image repeating the last value
              on the edge. For example, padding [1, 2, 3, 4] with 2 elements on
              both sides in symmetric mode will result in
              [2, 1, 1, 2, 3, 4, 4, 3]

    Returns:
        ndarray: The padded image.
    """

    assert (shape is not None) ^ (padding is not None)
    if shape is not None:
        width = max(shape[1] - img.shape[1], 0)
        height = max(shape[0] - img.shape[0], 0)
        padding = (0, 0, width, height)

    # check pad_val
    if isinstance(pad_val, tuple):
        assert len(pad_val) == img.shape[-1]
    elif not isinstance(pad_val, numbers.Number):
        raise TypeError('pad_val must be a int or a tuple. '
                        f'But received {type(pad_val)}')

    # check padding
    if isinstance(padding, tuple) and len(padding) in [2, 4]:
        if len(padding) == 2:
            padding = (padding[0], padding[1], padding[0], padding[1])
    elif isinstance(padding, numbers.Number):
        padding = (padding, padding, padding, padding)
    else:
        raise ValueError('Padding must be a int or a 2, or 4 element tuple.'
                         f'But received {padding}')

    # check padding mode
    assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric']

    border_type = {
        'constant': cv2.BORDER_CONSTANT,
        'edge': cv2.BORDER_REPLICATE,
        'reflect': cv2.BORDER_REFLECT_101,
        'symmetric': cv2.BORDER_REFLECT
    }
    img = cv2.copyMakeBorder(
        img,
        padding[1],
        padding[3],
        padding[0],
        padding[2],
        border_type[padding_mode],
        value=pad_val)

    return img



@PIPELINES.register_module()
class IRandomResizedCrop(object):
    """Crop the given image to random size and aspect ratio.

    A crop of random size (default: of 0.08 to 1.0) of the original size and a
    random aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio
    is made. This crop is finally resized to given size.

    Args:
        size (sequence | int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        scale (tuple): Range of the random size of the cropped image compared
            to the original image. Defaults to (0.08, 1.0).
        ratio (tuple): Range of the random aspect ratio of the cropped image
            compared to the original image. Defaults to (3. / 4., 4. / 3.).
        max_attempts (int): Maximum number of attempts before falling back to
            Central Crop. Defaults to 10.
        efficientnet_style (bool): Whether to use efficientnet style Random
            ResizedCrop. Defaults to False.
        min_covered (Number): Minimum ratio of the cropped area to the original
             area. Only valid if efficientnet_style is true. Defaults to 0.1.
        crop_padding (int): The crop padding parameter in efficientnet style
            center crop. Only valid if efficientnet_style is true.
            Defaults to 32.
        interpolation (str): Interpolation method, accepted values are
            'nearest', 'bilinear', 'bicubic', 'area', 'lanczos'. Defaults to
            'bilinear'.
        backend (str): The image resize backend type, accepted values are
            `cv2` and `pillow`. Defaults to `cv2`.
    """

    interpolation_map={
        'nearest': cv2.INTER_NEAREST ,
        'bilinear': cv2.INTER_LINEAR ,
        'bicubic': cv2.INTER_CUBIC,
        'area': cv2.INTER_LINEAR,
        'lanczos': cv2.INTER_LANCZOS4
    }

    def __init__(self,
                 size,
                 scale=(0.08, 1.0),
                 ratio=(3. / 4., 4. / 3.),
                 max_attempts=10,
                 efficientnet_style=False,
                 min_covered=0.1,
                 crop_padding=32,
                 interpolation='bilinear'):
        if efficientnet_style:
            assert isinstance(size, int)
            self.size = (size, size)
            assert crop_padding >= 0
        else:
            if isinstance(size, (tuple, list)):
                self.size = size
            else:
                self.size = ((int)(size), (int)(size))
        if (scale[0] > scale[1]) or (ratio[0] > ratio[1]):
            raise ValueError('range should be of kind (min, max). '
                             f'But received scale {scale} and rato {ratio}.')
        assert min_covered >= 0, 'min_covered should be no less than 0.'
        assert isinstance(max_attempts, int) and max_attempts >= 0, \
            'max_attempts mush be int and no less than 0.'
        assert interpolation in ('nearest', 'bilinear', 'bicubic', 'area',
                                 'lanczos')

        self.scale = scale
        self.ratio = ratio
        self.max_attempts = max_attempts
        self.efficientnet_style = efficientnet_style
        self.min_covered = min_covered
        self.crop_padding = crop_padding
        self.interpolation = interpolation

    @staticmethod
    def get_params(img, scale, ratio, max_attempts=10):
        """Get parameters for ``crop`` for a random sized crop.

        Args:
            img (ndarray): Image to be cropped.
            scale (tuple): Range of the random size of the cropped image
                compared to the original image size.
            ratio (tuple): Range of the random aspect ratio of the cropped
                image compared to the original image area.
            max_attempts (int): Maximum number of attempts before falling back
                to central crop. Defaults to 10.

        Returns:
            tuple: Params (ymin, xmin, ymax, xmax) to be passed to `crop` for
                a random sized crop.
        """
        height = img.shape[0]
        width = img.shape[1]
        area = height * width

        for _ in range(max_attempts):
            target_area = random.uniform(*scale) * area
            log_ratio = (math.log(ratio[0]), math.log(ratio[1]))
            aspect_ratio = math.exp(random.uniform(*log_ratio))

            target_width = int(round(math.sqrt(target_area * aspect_ratio)))
            target_height = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < target_width <= width and 0 < target_height <= height:
                ymin = random.randint(0, height - target_height)
                xmin = random.randint(0, width - target_width)
                ymax = ymin + target_height - 1
                xmax = xmin + target_width - 1
                return ymin, xmin, ymax, xmax

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if in_ratio < min(ratio):
            target_width = width
            target_height = int(round(target_width / min(ratio)))
        elif in_ratio > max(ratio):
            target_height = height
            target_width = int(round(target_height * max(ratio)))
        else:  # whole image
            target_width = width
            target_height = height
        ymin = (height - target_height) // 2
        xmin = (width - target_width) // 2
        ymax = ymin + target_height - 1
        xmax = xmin + target_width - 1
        return ymin, xmin, ymax, xmax

    # https://github.com/kakaobrain/fast-autoaugment/blob/master/FastAutoAugment/data.py # noqa
    @staticmethod
    def get_params_efficientnet_style(img,
                                      size,
                                      scale,
                                      ratio,
                                      max_attempts=10,
                                      min_covered=0.1,
                                      crop_padding=32):
        """Get parameters for ``crop`` for a random sized crop in efficientnet
        style.

        Args:
            img (ndarray): Image to be cropped.
            size (sequence): Desired output size of the crop.
            scale (tuple): Range of the random size of the cropped image
                compared to the original image size.
            ratio (tuple): Range of the random aspect ratio of the cropped
                image compared to the original image area.
            max_attempts (int): Maximum number of attempts before falling back
                to central crop. Defaults to 10.
            min_covered (Number): Minimum ratio of the cropped area to the
                original area. Only valid if efficientnet_style is true.
                Defaults to 0.1.
            crop_padding (int): The crop padding parameter in efficientnet
                style center crop. Defaults to 32.

        Returns:
            tuple: Params (ymin, xmin, ymax, xmax) to be passed to `crop` for
                a random sized crop.
        """
        height, width = img.shape[:2]
        area = height * width
        min_target_area = scale[0] * area
        max_target_area = scale[1] * area

        for _ in range(max_attempts):
            aspect_ratio = random.uniform(*ratio)
            min_target_height = int(
                round(math.sqrt(min_target_area / aspect_ratio)))
            max_target_height = int(
                round(math.sqrt(max_target_area / aspect_ratio)))

            if max_target_height * aspect_ratio > width:
                max_target_height = int((width + 0.5 - 1e-7) / aspect_ratio)
                if max_target_height * aspect_ratio > width:
                    max_target_height -= 1

            max_target_height = min(max_target_height, height)
            min_target_height = min(max_target_height, min_target_height)

            # slightly differs from tf implementation
            target_height = int(
                round(random.uniform(min_target_height, max_target_height)))
            target_width = int(round(target_height * aspect_ratio))
            target_area = target_height * target_width

            # slight differs from tf. In tf, if target_area > max_target_area,
            # area will be recalculated
            if (target_area < min_target_area or target_area > max_target_area
                    or target_width > width or target_height > height
                    or target_area < min_covered * area):
                continue

            ymin = random.randint(0, height - target_height)
            xmin = random.randint(0, width - target_width)
            ymax = ymin + target_height - 1
            xmax = xmin + target_width - 1

            return ymin, xmin, ymax, xmax

        # Fallback to central crop
        img_short = min(height, width)
        crop_size = size[0] / (size[0] + crop_padding) * img_short

        ymin = max(0, int(round((height - crop_size) / 2.)))
        xmin = max(0, int(round((width - crop_size) / 2.)))
        ymax = min(height, ymin + crop_size) - 1
        xmax = min(width, xmin + crop_size) - 1

        return ymin, xmin, ymax, xmax

    def __call__(self, results):
        for key in ['image']:
            img = results[key]
            if self.efficientnet_style:
                get_params_func = self.get_params_efficientnet_style
                get_params_args = dict(
                    img=img,
                    size=self.size,
                    scale=self.scale,
                    ratio=self.ratio,
                    max_attempts=self.max_attempts,
                    min_covered=self.min_covered,
                    crop_padding=self.crop_padding)
            else:
                get_params_func = self.get_params
                get_params_args = dict(
                    img=img,
                    scale=self.scale,
                    ratio=self.ratio,
                    max_attempts=self.max_attempts)
            ymin, xmin, ymax, xmax = get_params_func(**get_params_args)
            img = imcrop(img, bboxes=np.array([xmin, ymin, xmax, ymax]))
            results[key] = cv2.resize(
                img,
                tuple(self.size[::-1]),
                interpolation=RandomResizedCrop.interpolation_map[self.interpolation])
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__ + f'(size={self.size}'
        repr_str += f', scale={tuple(round(s, 4) for s in self.scale)}'
        repr_str += f', ratio={tuple(round(r, 4) for r in self.ratio)}'
        repr_str += f', max_attempts={self.max_attempts}'
        repr_str += f', efficientnet_style={self.efficientnet_style}'
        repr_str += f', min_covered={self.min_covered}'
        repr_str += f', crop_padding={self.crop_padding}'
        repr_str += f', interpolation={self.interpolation}'
        return repr_str

@PIPELINES.register_module()
class IRandomFlip(object):
    """Flip the image randomly.

    Flip the image randomly based on flip probaility and flip direction.

    Args:
        flip_prob (float): probability of the image being flipped. Default: 0.5
        direction (str): The flipping direction. Options are
            'horizontal' and 'vertical'. Default: 'horizontal'.
    """

    def __init__(self, flip_prob=0.5, direction='horizontal'):
        assert 0 <= flip_prob <= 1
        assert direction in ['horizontal', 'vertical']
        self.flip_prob = flip_prob
        self.direction = direction

    def __call__(self, results):
        """Call function to flip image.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Flipped results, 'flip', 'flip_direction' keys are added into
                result dict.
        """
        flip = True if np.random.rand() < self.flip_prob else False
        results['flip'] = flip
        results['flip_direction'] = self.direction
        if results['flip']:
            # flip image
            for key in ['image']:
                results[key] = imflip(
                    results[key], direction=results['flip_direction'])
        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(flip_prob={self.flip_prob})'


@PIPELINES.register_module()
class IRandomErasing(object):
    """Randomly selects a rectangle region in an image and erase pixels.

    Args:
        erase_prob (float): Probability that image will be randomly erased.
            Default: 0.5
        min_area_ratio (float): Minimum erased area / input image area
            Default: 0.02
        max_area_ratio (float): Maximum erased area / input image area
            Default: 0.4
        aspect_range (sequence | float): Aspect ratio range of erased area.
            if float, it will be converted to (aspect_ratio, 1/aspect_ratio)
            Default: (3/10, 10/3)
        mode (str): Fill method in erased area, can be:

            - const (default): All pixels are assign with the same value.
            - rand: each pixel is assigned with a random value in [0, 255]

        fill_color (sequence | Number): Base color filled in erased area.
            Defaults to (128, 128, 128).
        fill_std (sequence | Number, optional): If set and ``mode`` is 'rand',
            fill erased area with random color from normal distribution
            (mean=fill_color, std=fill_std); If not set, fill erased area with
            random color from uniform distribution (0~255). Defaults to None.

    Note:
        See `Random Erasing Data Augmentation
        <https://arxiv.org/pdf/1708.04896.pdf>`_

        This paper provided 4 modes: RE-R, RE-M, RE-0, RE-255, and use RE-M as
        default. The config of these 4 modes are:

        - RE-R: RandomErasing(mode='rand')
        - RE-M: RandomErasing(mode='const', fill_color=(123.67, 116.3, 103.5))
        - RE-0: RandomErasing(mode='const', fill_color=0)
        - RE-255: RandomErasing(mode='const', fill_color=255)
    """

    def __init__(self,
                 erase_prob=0.5,
                 min_area_ratio=0.02,
                 max_area_ratio=0.4,
                 aspect_range=(3 / 10, 10 / 3),
                 mode='const',
                 fill_color=(128, 128, 128),
                 fill_std=None):
        assert isinstance(erase_prob, float) and 0. <= erase_prob <= 1.
        assert isinstance(min_area_ratio, float) and 0. <= min_area_ratio <= 1.
        assert isinstance(max_area_ratio, float) and 0. <= max_area_ratio <= 1.
        assert min_area_ratio <= max_area_ratio, \
            'min_area_ratio should be smaller than max_area_ratio'
        if isinstance(aspect_range, float):
            aspect_range = min(aspect_range, 1 / aspect_range)
            aspect_range = (aspect_range, 1 / aspect_range)
        assert isinstance(aspect_range, Sequence) and len(aspect_range) == 2 \
            and all(isinstance(x, float) for x in aspect_range), \
            'aspect_range should be a float or Sequence with two float.'
        assert all(x > 0 for x in aspect_range), \
            'aspect_range should be positive.'
        assert aspect_range[0] <= aspect_range[1], \
            'In aspect_range (min, max), min should be smaller than max.'
        assert mode in ['const', 'rand']
        if isinstance(fill_color, Number):
            fill_color = [fill_color] * 3
        assert isinstance(fill_color, Sequence) and len(fill_color) == 3 \
            and all(isinstance(x, Number) for x in fill_color), \
            'fill_color should be a float or Sequence with three int.'
        if fill_std is not None:
            if isinstance(fill_std, Number):
                fill_std = [fill_std] * 3
            assert isinstance(fill_std, Sequence) and len(fill_std) == 3 \
                and all(isinstance(x, Number) for x in fill_std), \
                'fill_std should be a float or Sequence with three int.'

        self.erase_prob = erase_prob
        self.min_area_ratio = min_area_ratio
        self.max_area_ratio = max_area_ratio
        self.aspect_range = aspect_range
        self.mode = mode
        self.fill_color = fill_color
        self.fill_std = fill_std

    def _fill_pixels(self, img, top, left, h, w):
        if self.mode == 'const':
            patch = np.empty((h, w, 3), dtype=np.uint8)
            patch[:, :] = np.array(self.fill_color, dtype=np.uint8)
        elif self.fill_std is None:
            # Uniform distribution
            patch = np.random.uniform(0, 256, (h, w, 3)).astype(np.uint8)
        else:
            # Normal distribution
            patch = np.random.normal(self.fill_color, self.fill_std, (h, w, 3))
            patch = np.clip(patch.astype(np.int32), 0, 255).astype(np.uint8)

        img[top:top + h, left:left + w] = patch
        return img

    def __call__(self, results):
        """
        Args:
            results (dict): Results dict from pipeline

        Returns:
            dict: Results after the transformation.
        """
        for key in ['image']:
            if np.random.rand() > self.erase_prob:
                continue
            img = results[key]
            img_h, img_w = img.shape[:2]

            # convert to log aspect to ensure equal probability of aspect ratio
            log_aspect_range = np.log(
                np.array(self.aspect_range, dtype=np.float32))
            aspect_ratio = np.exp(np.random.uniform(*log_aspect_range))
            area = img_h * img_w
            area *= np.random.uniform(self.min_area_ratio, self.max_area_ratio)

            h = min(int(round(np.sqrt(area * aspect_ratio))), img_h)
            w = min(int(round(np.sqrt(area / aspect_ratio))), img_w)
            top = np.random.randint(0, img_h - h) if img_h > h else 0
            left = np.random.randint(0, img_w - w) if img_w > w else 0
            img = self._fill_pixels(img, top, left, h, w)

            results[key] = img
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(erase_prob={self.erase_prob}, '
        repr_str += f'min_area_ratio={self.min_area_ratio}, '
        repr_str += f'max_area_ratio={self.max_area_ratio}, '
        repr_str += f'aspect_range={self.aspect_range}, '
        repr_str += f'mode={self.mode}, '
        repr_str += f'fill_color={self.fill_color}, '
        repr_str += f'fill_std={self.fill_std})'
        return repr_str


def imnormalize(img, mean, std, to_rgb=True):
    """Normalize an image with mean and std.

    Args:
        img (ndarray): Image to be normalized.
        mean (ndarray): The mean to be used for normalize.
        std (ndarray): The std to be used for normalize.
        to_rgb (bool): Whether to convert to rgb.

    Returns:
        ndarray: The normalized image.
    """
    img = img.copy().astype(np.float32)
    return imnormalize_(img, mean, std, to_rgb)


def imnormalize_(img, mean, std, to_rgb=True):
    """Inplace normalize an image with mean and std.

    Args:
        img (ndarray): Image to be normalized.
        mean (ndarray): The mean to be used for normalize.
        std (ndarray): The std to be used for normalize.
        to_rgb (bool): Whether to convert to rgb.

    Returns:
        ndarray: The normalized image.
    """
    # cv2 inplace normalization does not accept uint8
    assert img.dtype != np.uint8
    mean = np.float64(mean.reshape(1, -1))
    stdinv = 1 / np.float64(std.reshape(1, -1))
    if to_rgb:
        cv2.cvtColor(img, cv2.COLOR_BGR2RGB, img)  # inplace
    cv2.subtract(img, mean, img)  # inplace
    cv2.multiply(img, stdinv, img)  # inplace
    return img



@PIPELINES.register_module()
class INormalize(object):
    """Normalize the image.

    Args:
        mean (sequence): Mean values of 3 channels.
        std (sequence): Std values of 3 channels.
        to_rgb (bool): Whether to convert the image from BGR to RGB,
            default is true.
    """

    def __init__(self, mean, std, to_rgb=False, keys=['image']):
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)
        self.to_rgb = to_rgb
        self.keys = keys

    def __call__(self, results):
        for key in self.keys:
            results[key] = imnormalize(results[key], self.mean, self.std,
                                            self.to_rgb)

        if 'image_meta' in results:
            results['image_meta']['img_norm_cfg'] = \
                dict(mean=self.mean, std=self.std, to_rgb=self.to_rgb)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(mean={list(self.mean)}, '
        repr_str += f'std={list(self.std)}, '
        repr_str += f'to_rgb={self.to_rgb})'
        return repr_str


@PIPELINES.register_module()
class IToGray(object):
    def __init__(self, keys=['image']):
        self.keys = keys

    def __call__(self, results):
        for key in self.keys:
            image = results[key]
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            results[key] = np.expand_dims(gray, -1)
        return results


@PIPELINES.register_module()
class INumpyToPIL(object):
    def __init__(self, keys=['image'], mode=None):
        self.keys = keys
        if mode is None:
            mode = 'RGB'
        assert(mode in ['RGB', 'L'])
        self.mode = mode

    def __call__(self, samples):
        for key in self.keys:
            samples[key] = Image.fromarray(samples[key]).convert(self.mode)
                
        return samples


@PIPELINES.register_module()
class IPILToNumpy(object):
    def __init__(self, keys=['image']):
        self.keys = keys

    def __call__(self, samples):
        for key in self.keys:
            samples[key] = np.array(samples[key])
                
        return samples


from __future__ import division
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import logging
import os
import typing
import torch.utils.data
import torchvision.transforms as transforms
import numpy as np
import copy
from antgo.framework.helper.utils import build_from_cfg


def register(cls):
    class ProxyReader(Reader):
        def __init__(self, train_or_test, dir='', params={}, pipeline=None, inputs_def=None, **kwargs):
            super().__init__(
                cls(train_or_test, dir, params),
                pipeline=pipeline,
                inputs_def=inputs_def,
            )
    from antgo.framework.helper.dataset.builder import DATASETS
    DATASETS.register_module(name=cls.__name__)(ProxyReader)   
    return ProxyReader


class KVReaderBase(torch.utils.data.Dataset):
    def __init__(self, pipeline=None, weak_pipeline=None, strong_pipeline=None) -> None:
        super().__init__()
        self.pipeline = []
        self.weak_pipeline = []
        self.strong_pipeline = []
        self.is_kv = True
        self.keys = []
        if pipeline is not None:
            from antgo.framework.helper.dataset import PIPELINES
            for transform in pipeline:
                if isinstance(transform, dict):
                    transform = build_from_cfg(transform, PIPELINES)
                    self.pipeline.append(transform)
                else:
                    raise TypeError('pipeline must be a dict')

            if weak_pipeline is not None and strong_pipeline is not None:
                for transform in weak_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.weak_pipeline.append(transform)
                    else:
                        raise TypeError('weak_pipeline must be a dict')
                
                for transform in strong_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.strong_pipeline.append(transform)
                    else:
                        raise TypeError('strong_pipeline must be a dict')
    
    def __len__(self):
        return len(self.keys)

    def reads(self, index):
        raise NotImplementedError

    def __getitem__(self, index):
        assert(isinstance(index, list))
        sample_list = []
        for sample in self.reads(index):
            weak_sample = None
            strong_sample = None
            if len(self.weak_pipeline) > 0 or len(self.strong_pipeline) > 0:
                weak_sample = copy.deepcopy(sample)
                for transform in self.weak_pipeline:
                    weak_sample = transform(weak_sample)

                strong_sample = copy.deepcopy(weak_sample)
                for transform in self.strong_pipeline:
                    strong_sample = transform(strong_sample)

            if weak_sample is not None and strong_sample is not None:
                for transform in self.pipeline:
                    weak_sample = transform(weak_sample)

                for transform in self.pipeline:
                    strong_sample = transform(strong_sample)

                sample_list.append(weak_sample)
                sample_list.append(strong_sample)
            else:
                for transform in self.pipeline:
                    sample = transform(sample)
            
                sample_list.append(sample)
        return sample_list


class Reader(torch.utils.data.Dataset):
    def __init__(self, dataset, pipeline=None, weak_pipeline=None, strong_pipeline=None, inputs_def=None):
        self.proxy_dataset = dataset
        self.pipeline = []
        self.pipeline_types = []
        self.weak_pipeline = []
        self.strong_pipeline = []        
        if pipeline is not None:
            from antgo.framework.helper.dataset import PIPELINES
            for transform in pipeline:
                if isinstance(transform, dict):
                    self.pipeline_types.append(transform['type'])
                    transform = build_from_cfg(transform, PIPELINES)
                    self.pipeline.append(transform)
                else:
                    raise TypeError('pipeline must be a dict')

            if weak_pipeline is not None and strong_pipeline is not None:
                for transform in weak_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.weak_pipeline.append(transform)
                    else:
                        raise TypeError('weak_pipeline must be a dict')
                
                for transform in strong_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.strong_pipeline.append(transform)
                    else:
                        raise TypeError('strong_pipeline must be a dict')

        self._fields = copy.deepcopy(inputs_def['fields']) if inputs_def else None
        self._alias = None
        if self._fields is not None and 'alias' in inputs_def:
            self._alias = copy.deepcopy(inputs_def['alias'])
        
        if self._fields is not None:
            if self._alias is None:
                self._alias = copy.deepcopy(self._fields)
                
        self.flag = np.zeros(len(self), dtype=np.uint8)
        if hasattr(self.proxy_dataset, 'flag'):
            self.flag = self.proxy_dataset.flag

        self.CLASSES = 1
        if hasattr(self.proxy_dataset, 'CLASSES'):
            self.CLASSES = self.proxy_dataset.CLASSES

    def _arrange(self, sample, fields, alias):
        if fields is None:
            return sample      
          
        if type(fields[0]) == list or type(fields[0]) == tuple:
            warp_ins = []
            for alia, field in zip(alias, fields):
                one_ins = {}
                for aa, ff in zip(alia, field):
                    one_ins[aa] = sample[ff]
                
                warp_ins.append(one_ins)
            return warp_ins
        
        warp_ins = {}
        for alia, field in zip(alias, fields):
            warp_ins[alia] = sample[field]

        return warp_ins

    def __len__(self):
        return self.proxy_dataset.size

    def __getitem__(self, idx):
        sample = None
        try:
            sample = self.proxy_dataset.sample(idx)
            if sample is None:
                fail_count = 0
                while True:
                    # 随机选择，找到有效样本
                    random_i = np.random.randint(0,self.proxy_dataset.size)
                    sample = self.proxy_dataset.sample(random_i)
                    if sample is not None:
                        break

                    fail_count += 1
                    if fail_count > 10:
                        logging.warn(f'Fail find correct sample and exceed count {fail_count}.')
        except:
            print(f'sample error {idx}')

        weak_sample = None
        strong_sample = None
        if len(self.weak_pipeline) > 0 or len(self.strong_pipeline) > 0:
            weak_sample = copy.deepcopy(sample)
            for transform in self.weak_pipeline:
                weak_sample = transform(weak_sample)
            strong_sample = copy.deepcopy(weak_sample)
            for transform in self.strong_pipeline:
                strong_sample = transform(strong_sample)

        if weak_sample is not None and strong_sample is not None:
            for transform in self.pipeline:
                weak_sample = transform(weak_sample)

            for transform in self.pipeline:
                strong_sample = transform(strong_sample)

            # arange warp
            weak_sample = self._arrange(weak_sample, self._fields, self._alias)
            strong_sample = self._arrange(strong_sample, self._fields, self._alias)
            return [weak_sample, strong_sample]
        else:    
            for transform in self.pipeline:
                sample = transform(sample)

            # arange warp
            sample = self._arrange(sample, self._fields, self._alias)
            return sample

    def get_cat_ids(self, idx):
        return self.proxy_dataset.get_cat_ids(idx)

    def get_ann_info(self, idx):
        return self.proxy_dataset.get_ann_info(idx)

    def evaluate(self, preds,**kwargs):
        return self.proxy_dataset.evaluate(preds, **kwargs)


class TVReader(torch.utils.data.Dataset):
    def __init__(self, dataset, pipeline=None, weak_pipeline=None, strong_pipeline=None, inputs_def=None):
        self.proxy_dataset = dataset
        self.pipeline = []
        self.pipeline_types = []
        self.weak_pipeline = []
        self.strong_pipeline = []        
        if pipeline is not None:
            from antgo.framework.helper.dataset import PIPELINES
            for transform in pipeline:
                if isinstance(transform, dict):
                    self.pipeline_types.append(transform['type'])
                    transform = build_from_cfg(transform, PIPELINES)
                    self.pipeline.append(transform)
                else:
                    raise TypeError('pipeline must be a dict')

            if weak_pipeline is not None and strong_pipeline is not None:
                for transform in weak_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.weak_pipeline.append(transform)
                    else:
                        raise TypeError('weak_pipeline must be a dict')
                
                for transform in strong_pipeline:
                    if isinstance(transform, dict):
                        transform = build_from_cfg(transform, PIPELINES)
                        self.strong_pipeline.append(transform)
                    else:
                        raise TypeError('strong_pipeline must be a dict')

        self._fields = copy.deepcopy(inputs_def['fields']) if inputs_def else None
        self._alias = None
        if self._fields is not None and 'alias' in inputs_def:
            self._alias = copy.deepcopy(inputs_def['alias'])
        
        if self._fields is not None:
            if self._alias is None:
                self._alias = copy.deepcopy(self._fields)

        self.flag = np.zeros(len(self), dtype=np.uint8)
        if hasattr(self.proxy_dataset, 'flag'):
            self.flag = self.proxy_dataset.flag

        self.CLASSES = 1
        if hasattr(self.proxy_dataset, 'CLASSES'):
            self.CLASSES = self.proxy_dataset.CLASSES

    def _arrange(self, sample, fields, alias):
        if fields is None:
            return sample      
          
        if type(fields[0]) == list or type(fields[0]) == tuple:
            warp_ins = []
            for alia, field in zip(alias, fields):
                one_ins = {}
                for aa, ff in zip(alia, field):
                    one_ins[aa] = sample[ff]
                
                warp_ins.append(one_ins)
            return warp_ins
        
        warp_ins = {}
        for alia, field in zip(alias, fields):
            warp_ins[alia] = sample[field]

        return warp_ins
    
    def __len__(self):
        return len(self.proxy_dataset)
    
    def __getitem__(self, idx):
        sample = None
        try:
            sample = self.proxy_dataset[idx]
            if sample is None:
                fail_count = 0
                while True:
                    # 随机选择，找到有效样本
                    random_i = np.random.randint(0,self.proxy_dataset.size)
                    sample = self.proxy_dataset.sample(random_i)
                    if sample is not None:
                        break

                    fail_count += 1
                    if fail_count > 10:
                        logging.warn(f'Fail find correct sample and exceed count {fail_count}.')            

            if isinstance(sample, tuple) or isinstance(sample, list):
                temp = {}
                for data, name in zip(sample, self._alias):
                    temp[name] = data
                sample = temp
            else:
                sample = {self._alias[0]: sample}
        except:
            print(f'sample error {idx}')
        
        weak_sample = None
        strong_sample = None
        if len(self.weak_pipeline) > 0 or len(self.strong_pipeline) > 0:
            weak_sample = copy.deepcopy(sample)
            for transform in self.weak_pipeline:
                weak_sample = transform(weak_sample)
            strong_sample = copy.deepcopy(weak_sample)
            for transform in self.strong_pipeline:
                strong_sample = transform(strong_sample)

        if weak_sample is not None and strong_sample is not None:
            for transform in self.pipeline:
                weak_sample = transform(weak_sample)

            for transform in self.pipeline:
                strong_sample = transform(strong_sample)

            # arange warp
            weak_sample = self._arrange(weak_sample, self._fields, self._alias)
            strong_sample = self._arrange(strong_sample, self._fields, self._alias)
            return [weak_sample, strong_sample]
        else:
            for transform in self.pipeline:
                sample = transform(sample)

            # arange warp
            sample = self._arrange(sample, self._fields, self._alias)
            return sample



from __future__ import unicode_literals
from __future__ import division

import numpy as np
import unittest
import json


def tied_rank(x):
    """
    Computes the tied rank of elements in x.
    This function computes the tied rank of elements in x.
    Parameters
    ----------
    x : list of numbers, numpy array
    Returns
    -------
    score : list of numbers
            The tied rank f each element in x
    """
    sorted_x = sorted(zip(x, range(len(x))))
    r = [0 for k in x]

    cur_val = sorted_x[0][0]
    offset = 0
    for i in range(len(sorted_x)):
        if sorted_x[i][0] > -float('inf'):
            cur_val = sorted_x[i][0]
            offset = i
            break

    last_rank = 0
    for i in range(len(sorted_x)):
        if sorted_x[i][0] <= -float('inf'):
            continue

        if cur_val != sorted_x[i][0]:
            cur_val = sorted_x[i][0]
            ioffset = i - offset
            for j in range(last_rank, ioffset):
                r[sorted_x[j + offset][1]] = float(last_rank+1+ioffset)/2.0
            last_rank = ioffset

        if i == len(sorted_x)-1:
            ioffset = i - offset
            for j in range(last_rank, ioffset+1):
                r[sorted_x[j + offset][1]] = float(last_rank+ioffset+2)/2.0
    return r


def auc(actual, posterior):
    """
    Computes the area under the receiver-operater characteristic (AUC) (only for binary classification task)
    References: http://cs.ru.nl/~tomh/onderwijs/dm/dm_files/roc_auc.pdf
    Parameters
    ----------
    actual : list of binary numbers, numpy array
             The ground truth value
    posterior : same type as actual
                Defines a ranking on the binary numbers, from most likely to
                be positive to least likely to be positive.
    Returns
    -------
    score : double
            The mean squared error between actual and posterior
    """
    # transform to array
    if type(actual) == list:
        actual = np.array(actual)
    if type(posterior) == list:
        posterior = np.array(posterior)

    if actual.shape[0] == 0 or posterior.shape[0] == 0:
        return None

    # transform to array(int)
    if actual.dtype != int:
        actual = actual.astype(dtype=int)

    # absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(posterior.shape) > 1 and posterior.shape[1] > 1:
        posterior = posterior[:, 0]
    posterior = np.reshape(posterior, [posterior.size])

    r = tied_rank(posterior.tolist())
    det_num = len(np.where(posterior > -float('inf'))[0])
    det_positive_index = [i for i in range(len(actual)) if actual[i] == 1 and posterior[i] > -float('inf')]
    det_num_positive = len(det_positive_index)
    num_positive = len(np.where(actual == 1)[0])    # detected positive and missed positive
    det_num_negative = det_num - det_num_positive

    det_sum_positive = sum([r[i] for i in det_positive_index])
    auc = ((det_sum_positive - det_num_positive * (det_num_positive + 1)/2.0) /
           (det_num_negative * num_positive + 0.000001))
    return auc


def roc(actual,posterior):
    '''
    compute receiver-operater characteristic curve(ROC) for binary classification
    Parameters
    ----------
    actual: list
            ground truth label
    posterior: list
            predicated score for positive label

    Returns
    -------
    x: false positive rate
    y: true positive rate
    '''
    # transform to array
    if type(actual) == list:
        actual = np.array(actual)
    if type(posterior) == list:
        posterior = np.array(posterior)

    if actual.shape[0] == 0 or posterior.shape[0] == 0:
        return None

    # transform to array(int)
    if actual.dtype != int:
        actual = actual.astype(dtype=int)

    # absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(posterior.shape) > 1 and posterior.shape[1] > 1:
        posterior = posterior[:, 0]
    posterior = np.reshape(posterior, [posterior.size])

    #
    num = actual.shape[0]
    positive_num = len(np.where(actual == 1)[0])
    negative_num = num - positive_num

    true_positive_num = 0.0
    false_positive_num = 0.0

    sorted_x = sorted(zip(posterior, range(len(posterior))), reverse=True)
    fp_thres = np.array(range(11)) * 0.1

    fp = []
    tp = []

    for i in range(len(sorted_x)):
        if posterior[sorted_x[i][1]] <= -float("inf"):
            continue

        if actual[sorted_x[i][1]] == 1:
            true_positive_num += 1.0
        else:
            false_positive_num += 1.0

        # false positive rate
        negative_num = negative_num if negative_num > 0 else negative_num + 0.0000000001
        fp_rate = false_positive_num / float(negative_num)
        # true positive rate
        positive_num = positive_num if positive_num > 0 else positive_num + 0.0000000001
        tp_rate = true_positive_num / float(positive_num)

        fp.append(fp_rate)
        tp.append(tp_rate)

    inds = np.searchsorted(fp, fp_thres, side='left')
    x = [0 for _ in range(len(inds))]
    y = [0 for _ in range(len(inds))]
    try:
        for ri, pi in enumerate(inds):
            if pi == len(tp):
                y[ri] = tp[pi - 1]
                x[ri] = fp[pi - 1]
            else:
                y[ri] = tp[pi]
                x[ri] = fp[pi]
    except:
        pass

    xy = np.array([x, y])
    return xy.transpose()



import warnings

import torch.nn as nn
from antgo.framework.helper.cnn import ConvModule, DepthwiseSeparableConvModule
from antgo.framework.helper.runner import BaseModule
from torch.nn.modules.batchnorm import _BatchNorm
import torch.nn.functional as F
from antgo.framework.helper.models.builder import BACKBONES
from antgo.framework.helper.cnn.bricks.inverted_residual import InvertedResidual
from antgo.framework.helper.cnn.bricks.make_divisible import make_divisible
import torch


@BACKBONES.register_module()
class DDRMobileNetV2(BaseModule):
    """MobileNetV2 backbone.

    Args:
        widen_factor (float): Width multiplier, multiply number of
            channels in each layer by this amount. Default: 1.0.
        out_indices (Sequence[int], optional): Output from which stages.
            Default: (1, 2, 4, 7).
        frozen_stages (int): Stages to be frozen (all param fixed).
            Default: -1, which means not freezing any parameters.
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU6').
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        pretrained (str, optional): model pretrained path. Default: None
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None
    """

    # Parameters to build layers. 4 parameters are needed to construct a
    # layer, from left to right: expand_ratio, channel, num_blocks, stride.
    arch_settings = [[1, 16, 1, 1], [6, 24, 2, 2], [6, 32, 3, 2],
                     [6, 64, 4, 2], [6, 96, 3, 1], [6, 160, 3, 2],
                     [6, 320, 1, 1]]

    def __init__(self,
                 in_channels=1,
                 widen_factor=1.,
                 out_indices=(1, 2, 4, 6),
                 frozen_stages=-1,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 norm_eval=False,
                 with_cp=False,
                 pretrained=None,
                 init_cfg=None, **kwargs):
        super(DDRMobileNetV2, self).__init__(init_cfg)

        self.pretrained = pretrained
        assert not (init_cfg and pretrained), \
            'init_cfg and pretrained cannot be specified at the same time'
        if isinstance(pretrained, str):
            warnings.warn('DeprecationWarning: pretrained is deprecated, '
                          'please use "init_cfg" instead')
            self.init_cfg = dict(type='Pretrained', checkpoint=pretrained)
        elif pretrained is None:
            if init_cfg is None:
                self.init_cfg = [
                    dict(type='Kaiming', layer='Conv2d'),
                    dict(
                        type='Constant',
                        val=1,
                        layer=['_BatchNorm', 'GroupNorm'])
                ]
        else:
            raise TypeError('pretrained must be a str or None')

        self.widen_factor = widen_factor
        self.out_indices = out_indices
        if not set(out_indices).issubset(set(range(0, 8))):
            raise ValueError('out_indices must be a subset of range'
                             f'(0, 8). But received {out_indices}')

        if frozen_stages not in range(-1, 8):
            raise ValueError('frozen_stages must be in range(-1, 8). '
                             f'But received {frozen_stages}')
        self.out_indices = out_indices
        self.frozen_stages = frozen_stages
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.norm_eval = norm_eval
        self.with_cp = with_cp

        self.in_channels = make_divisible(32 * widen_factor, 8)

        self.conv1 = ConvModule(
            in_channels=in_channels,
            out_channels=self.in_channels,
            kernel_size=3,
            stride=2,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

        self.layers = []
        self.channels_num = {}
        for i, layer_cfg in enumerate(self.arch_settings):
            expand_ratio, channel, num_blocks, stride = layer_cfg
            out_channels = make_divisible(channel * widen_factor, 8)
            inverted_res_layer = self.make_layer(
                out_channels=out_channels,
                num_blocks=num_blocks,
                stride=stride,
                expand_ratio=expand_ratio,
                in_channels= self.in_channels * 2 if i in [5] else None
            )
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, inverted_res_layer)
            self.layers.append(layer_name)
            self.channels_num[i] = out_channels
        
        # 编码通道数
        self.encoder_ch = 64
        
        # 1,    2,      4,      6
        # 1/4,  1/8,    1/16,   1/32
        self.down4_to_16 = \
                ConvModule(
                    self.encoder_ch, 
                    self.channels_num[4],
                    kernel_size=3,
                    stride=4, 
                    padding=(int)((4*(3-1)+1-4)//2+1), 
                    dilation=4, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        self.down4_to_32 = \
                ConvModule(
                    self.encoder_ch, 
                    self.channels_num[6],
                    kernel_size=3,
                    stride=8, 
                    padding=(int)((6*(3-1)+1-8)//2+1), 
                    dilation=6, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))             

        self.compression8_to_4 = \
                ConvModule(
                    self.channels_num[2], 
                    self.encoder_ch, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        self.compression16_to_4 = \
                ConvModule(
                    self.channels_num[4], 
                    self.encoder_ch, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        
        self.layer4_1 = DepthwiseSeparableConvModule(
            in_channels=self.channels_num[1],
            out_channels=self.encoder_ch,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )
        self.layer4_2 = DepthwiseSeparableConvModule(
            in_channels=self.encoder_ch,
            out_channels=self.encoder_ch,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )
        self.layer4_3 = DepthwiseSeparableConvModule(
            in_channels=self.encoder_ch,
            out_channels=self.encoder_ch,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )
        self.layer4_4 = DepthwiseSeparableConvModule(
            in_channels=self.encoder_ch,
            out_channels=self.encoder_ch,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )
        self.layer4_5 = DepthwiseSeparableConvModule(
            in_channels=self.encoder_ch+256,
            out_channels=self.encoder_ch,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )    
      
        # 1/4
        # self.encoder_enchance = WaterFall(self.encoder_ch, self.encoder_ch, self.encoder_ch)

        # 1/32
        self.enchance = \
            ConvModule(
                    self.channels_num[6]*2, 
                    256,
                    kernel_size=3,
                    stride=1, 
                    padding=1, 
                    act_cfg=None,
                    norm_cfg=dict(type='BN'))

    def make_layer(self, out_channels, num_blocks, stride, expand_ratio, in_channels=None):
        """Stack InvertedResidual blocks to build a layer for MobileNetV2.

        Args:
            out_channels (int): out_channels of block.
            num_blocks (int): number of blocks.
            stride (int): stride of the first block. Default: 1
            expand_ratio (int): Expand the number of channels of the
                hidden layer in InvertedResidual by this ratio. Default: 6.
        """
        layers = []
        for i in range(num_blocks):
            if i >= 1:
                stride = 1
            layers.append(
                InvertedResidual(
                    in_channels if (in_channels is not None and i == 0) else self.in_channels,
                    out_channels,
                    mid_channels=int(round(self.in_channels * expand_ratio)),
                    stride=stride,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg))
            self.in_channels = out_channels

        return nn.Sequential(*layers)

    def _freeze_stages(self):
        if self.frozen_stages >= 0:
            for param in self.conv1.parameters():
                param.requires_grad = False
        for i in range(1, self.frozen_stages + 1):
            layer = getattr(self, f'layer{i}')
            layer.eval()
            for param in layer.parameters():
                param.requires_grad = False

    def forward(self, x):
        """Forward function."""
        x = self.conv1(x)
        out_layer_1 = None
        out_layer_2 = None
        out_layer_3 = None
        for i, layer_name in enumerate(self.layers):
            layer = getattr(self, layer_name)
            x = layer(x)
            if i in self.out_indices:
                if i == 1:
                    # 1/4
                    out_layer_1 = self.layer4_1(x)
                elif i == 2:
                    # 1/8
                    # merge from 1/4
                    x_ = self.layer4_2(out_layer_1)       # channel 32
                    out_layer_2 = x_ + F.interpolate(
                                    self.compression8_to_4(x),
                                    scale_factor=2,
                                    mode='bilinear')
                    out_layer_2 = F.relu(out_layer_2)       # 1/4
                elif i == 4:
                    # 1/16
                    # merge from 1/4
                    # down_x = F.relu(x + self.down4_to_16(out_layer_2))
                    down_x = F.relu(torch.cat([x, self.down4_to_16(out_layer_2)], dim=1))
                    x_ = self.layer4_3(out_layer_2)
                    out_layer_3 = x_ + F.interpolate(
                                    self.compression16_to_4(x),
                                    scale_factor=4,
                                    mode='bilinear')
                    out_layer_3 = F.relu(out_layer_3)
                    x = down_x   
                elif i == 6:
                    # down_x = F.relu(x + self.down4_to_32(out_layer_3))
                    down_x = F.relu(torch.cat([x, self.down4_to_32(out_layer_3)], dim=1))
                    x = down_x

        # 在1/32分辨率，使用瀑布结果增强全局编码
        # 1/32 分辨率 4x4
        x = self.enchance(x)

        out_layer_4 = self.layer4_4(out_layer_3)
        out_layer = self.layer4_5(
            F.relu(torch.cat([out_layer_4, F.interpolate(x, scale_factor=8, mode='bilinear')], dim=1))
        )

        # 在1/4分辨率，使用瀑布结果抓取多视野信息
        # out_layer = self.encoder_enchance(out_layer)
        return tuple([out_layer, x])

    def init_weights(self):
        # pretrained = '/opt/tiger/handtt/mobilenet_v2_batch256_imagenet_20200708-3b2dc3af.pth'
        # # dd = self.load_state_dict(pretrained)
        # dd = torch.load(pretrained, map_location='cpu')['state_dict']

        # filter_state_dict = {}
        # for k, v in dd.items():
        #     filter_k = '.'.join(k.split('.')[1:])
        #     filter_state_dict[filter_k] = v

        # self.load_state_dict(filter_state_dict, strict=False)
        pass



from antgo.framework.helper.dataset.builder import DATASETS
from antgo.framework.helper import reader
from antgo.dataflow.dataset.dataset import Dataset
import numpy as np
import torch


# @reader.register
# class YourDataset(Dataset):
#     def __init__(self, train_or_test="train", dir=None, ext_params=None):
#         super().__init__(train_or_test, dir, ext_params)
#         pass
    
#     @property
#     def size(self):
#         # 返回数据集大小
#         return 0
    
#     def sample(self, id):
#         # 根据id，返回对应样本
#         return {}



from __future__ import unicode_literals
from __future__ import division

import numpy as np

def kdd_apk(actual, predicted, k=10):
    """
    Computes the average precision at k for Track 1 of the 2012 KDD Cup.
    This modified version uses the number of actual clicks as the denominator,
    regardless of k.
    This function computes the average prescision at k between two lists of
    items.
    Parameters
    ----------
    actual : list
             A list of elements that are to be predicted (order doesn't matter)
    predicted : list
                A list of predicted elements (order does matter)
    k : int, optional
        The maximum number of predicted elements
    Returns
    -------
    score : double
            The average precision at k over the input lists
    """
    if len(predicted)>k:
        predicted = predicted[:k]

    score = 0.0
    num_hits = 0.0

    for i,p in enumerate(predicted):
        if p in actual and p not in predicted[:i]:
            num_hits += 1.0
            score += num_hits / (i+1.0)

    if not actual:
        return 0.0

    return score / len(actual)

def kdd_mapk(actual, predicted, k=10):
    """
    Computes the mean average precision at k for Track 1 of the 2012 KDD Cup.
    This modified version uses the number of actual clicks as the denominator,
    regardless of k.
    This function computes the mean average prescision at k between two lists
    of lists of items.
    Parameters
    ----------
    actual : list
             A list of lists of elements that are to be predicted
             (order doesn't matter in the lists)
    predicted : list
                A list of lists of predicted elements
                (order matters in the lists)
    k : int, optional
        The maximum number of predicted elements
    Returns
    -------
    score : double
            The mean average precision at k over the input lists
    """
    return np.mean([kdd_apk(a,p,k) for a,p in zip(actual, predicted)])


# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
from PIL import Image


class GridMask(object):
    def __init__(self,
                 use_h=True,
                 use_w=True,
                 rotate=1,
                 offset=False,
                 ratio=0.5,
                 mode=1,
                 prob=0.7,
                 upper_iter=360000):
        super(GridMask, self).__init__()
        self.use_h = use_h
        self.use_w = use_w
        self.rotate = rotate
        self.offset = offset
        self.ratio = ratio
        self.mode = mode
        self.prob = prob
        self.st_prob = prob
        self.upper_iter = upper_iter

    def __call__(self, x, curr_iter):
        self.prob = self.st_prob * min(1, 1.0 * curr_iter / self.upper_iter)
        if np.random.rand() > self.prob:
            return x
        _, h, w = x.shape
        hh = int(1.5 * h)
        ww = int(1.5 * w)
        d = np.random.randint(2, h)
        self.l = min(max(int(d * self.ratio + 0.5), 1), d - 1)
        mask = np.ones((hh, ww), np.float32)
        st_h = np.random.randint(d)
        st_w = np.random.randint(d)
        if self.use_h:
            for i in range(hh // d):
                s = d * i + st_h
                t = min(s + self.l, hh)
                mask[s:t, :] *= 0
        if self.use_w:
            for i in range(ww // d):
                s = d * i + st_w
                t = min(s + self.l, ww)
                mask[:, s:t] *= 0

        r = np.random.randint(self.rotate)
        mask = Image.fromarray(np.uint8(mask))
        mask = mask.rotate(r)
        mask = np.asarray(mask)
        mask = mask[(hh - h) // 2:(hh - h) // 2 + h, (ww - w) // 2:(ww - w) // 2
                    + w].astype(np.float32)

        if self.mode == 1:
            mask = 1 - mask
        mask = np.expand_dims(mask, axis=0)
        if self.offset:
            offset = (2 * (np.random.rand(h, w) - 0.5)).astype(np.float32)
            x = (x * mask + offset * (1 - mask)).astype(x.dtype)
        else:
            x = (x * mask).astype(x.dtype)

        return x



import bisect
import collections
import copy
import math
from collections import defaultdict
from xml.etree.ElementTree import iselement

import numpy as np
from antgo.framework.helper.utils import build_from_cfg, print_log
import torch
from torch.utils.data.dataset import ConcatDataset as _ConcatDataset
from .builder import DATASETS, PIPELINES


@DATASETS.register_module()
class DatasetSamplingByClass(object):
    def __init__(self, dataset, sampling_num=4000, class_num=10) -> None:
        self.dataset = build_from_cfg(dataset, DATASETS, None)
        sample_labels = [self.dataset.get_cat_ids(i) for i in range(len(self.dataset))]
        label_per_class = sampling_num // class_num
        sample_labels = np.array(sample_labels)

        labeled_idx = []
        for i in range(class_num):
            idx = np.where(sample_labels == i)[0]
            idx = np.random.choice(idx, label_per_class, False)
            labeled_idx.extend(idx)
        labeled_idx = np.array(labeled_idx)
        assert len(labeled_idx) == sampling_num
        self.sampling_idx = labeled_idx

    def __getitem__(self, idx):
        return self.dataset[self.sampling_idx[idx]]

    def __len__(self):
        return len(self.sampling_idx)



# -*- coding: UTF-8 -*-
# @Time    : 2021/1/8 7:23 下午
# @File    : interactcontext.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.context import *
from antgo.ant.train import *
from antgo.ant.challenge import *
from antgo.ant.demo import *
from antgo.ant.browser import *
from antgo.ant.batch import *
from antgo.ant.ensemble import *
from antgo.ant.activelearning_v2 import *


class InteractContext(Context):
  def __init__(self):
    super(InteractContext, self).__init__(interact_mode=True)
    self.server_map = {}

  def _prepare_context(self, name, param, **kwargs):
    # 1.step 获取基本配置信息
    config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
    if not os.path.exists(config_xml):
      # 创建默认配置
      if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo')):
        os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo'))

      env = Environment(loader=FileSystemLoader('/'.join(os.path.realpath(__file__).split('/')[0:-1])))
      config_template = env.get_template('config.xml')
      config_data = {
        'DATASET_FACTORY': os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory'),
        'USER_TOKEN': ''
      }
      config_content = config_template.render(**config_data)
      with open(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'), 'w') as fp:
        fp.write(config_content)

      if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory')):
        os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory'))

      logger.warn('Build default antgo config in ~/.config/antgo/config.xml')

    Config.parse_xml(config_xml)
    factory = getattr(Config, 'factory', None)
    assert factory is not None

    if not os.path.exists(Config.data_factory):
      os.makedirs(Config.data_factory)
    if not os.path.exists(Config.task_factory):
      os.makedirs(Config.task_factory)

    self.name = name
    self.data_factory = Config.data_factory
    self.task_factory = Config.task_factory
    self.api_token = kwargs.get('token', Config.server_user_token)
    logger.info(f'Use experiment data_factory {self.data_factory}')
    logger.info(f'Use experiment task_factory {self.task_factory}')
    logger.info(f'Use experiment api token {self.api_token}')

    self.main_folder = os.path.abspath(os.curdir)
    if self.data_factory == os.path.join(os.environ['HOME'], '.config', 'antgo', 'factory'):
      # 当前正在使用默认配置
      logger.warn('Current use default factory, please shell "antgo config --dataset=xxx --token=xxx"')

    self.params = {}
    if type(param) == str:
      param = yaml.load(open(param, 'r'))

    if 'system' not in param:
      param.update({'system': {}})
    if 'research' not in param['system']:
      param['system']['research'] = ''
    if 'skip_training' not in param['system']:
      param['system']['skip_training'] = False
    if 'ip' not in param['system']:
      param['system']['ip'] = '127.0.0.1'
    if 'port' not in param['system']:
      param['system']['port'] = 8901
    if 'devices' not in param['system']:
      param['system']['devices'] = ''
    if 'running_platform' not in param['system']:
      param['system']['running_platform'] = ''
    if 'unlabel' not in param['system']:
      param['system']['unlabel'] = False
    if 'candidate' not in param['system']:
      param['system']['candidate'] = False
    if 'ext_params' not in param['system']:
      param['system']['ext_params'] = {}

    self.params = param
    # 2.step 设置来自实验名称
    self.dump_dir = os.path.join(os.path.abspath(os.curdir), 'dump')
    if not os.path.exists(self.dump_dir):
      os.makedirs(self.dump_dir)

    from_experiment = kwargs.get('from_experiment', None)
    if from_experiment is not None:
      experiment_path = os.path.join(self.dump_dir, from_experiment)
      if not os.path.exists(experiment_path):
        logger.error('Couldnt find experiment %s' % from_experiment)
        raise IOError

      if os.path.exists(os.path.join(experiment_path, 'train')):
        self.from_experiment = os.path.join(experiment_path, 'train')
      elif os.path.exists(os.path.join(experiment_path, 'inference')):
        self.from_experiment = os.path.join(experiment_path, 'inference')
      else:
        self.from_experiment = experiment_path

    # 3.step 设置默认任务
    task = kwargs.get('task', None)
    if task is not None:
      if os.path.exists(os.path.join(self.task_factory, task)):
        task = os.path.join(self.task_factory, task)
      else:
        task = os.path.join(self.main_folder, task)

    dataset = kwargs.get('dataset', 'unkown')
    if task is None and dataset is not None:
      # build default task
      with open(os.path.join(self.task_factory, '%s.xml' % name), 'w') as fp:
        task_content = '<task><task_name>%s</task_name><task_type>%s</task_type><task_badcase><badcase_num>%d</badcase_num><badcase_category>%d</badcase_category></task_badcase>' \
                       '<input>' \
                       '<source_data><data_set_name>%s</data_set_name>' \
                       '</source_data>' \
                       '<estimation_procedure><type>%s</type></estimation_procedure>' \
                       '<evaluation_measures><evaluation_measure>%s</evaluation_measure></evaluation_measures>' \
                       '</input>' \
                       '</task>' % (
                       name, '', 0, 0, dataset, '', '')
        fp.write(task_content)
      task = os.path.join(self.task_factory, '%s.xml' % name)

    self.dataset = dataset
    self.task = task

  @contextmanager
  def Train(self, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)
    # 2.step 创建时间戳
    time_stamp = timestamp()
    try:
      # 6.step 创建训练过程调度对象
      train = AntTrain(self,
                       exp_name,
                       self.data_factory,
                       self.dump_dir,
                       self.api_token,
                       self.task,
                       self.dataset,
                       time_stamp=time_stamp,
                       main_folder=self.main_folder)
      # 训练控制
      train.start()
      yield train

      # 日志更新
      mlogger.update()

      # 环境回收
      train.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Challenge(self, exp_name, exp_param, benchmark='', **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)

    try:
      # 5.step 创建挑战过程调度对象
      challenge = AntChallenge(self,
                               exp_name,
                               self.data_factory,
                               self.dump_dir,
                               self.api_token,
                               self.task,
                               benchmark,
                               time_stamp=timestamp(),
                               main_folder=self.main_folder)
      challenge.start()
      yield challenge

      # 日志更新
      mlogger.update()

      # 环境回收
      challenge.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Predict(self, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)

    if 'predict' in kwargs:
      params = self.params._params
      params.update({
        'predict': kwargs['predict']
      })
      self.params = params

    try:
      predict =\
        AntBatch(self,
                 exp_name,
                 self.api_token,
                 self.data_factory,
                 self.dump_dir,
                 self.task,
                 self.dataset)
      predict.start()
      yield predict

      if kwargs.get('clear', True):
        # 日志更新
        mlogger.update()

        # 环境回收
        predict.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Ensemble(self, exp_name, exp_param={}, **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)

    if 'config' in kwargs:
      params = self.params._params
      params.update({
        'ensemble': kwargs['config']
      })
      self.params = params

    try:
      ensemble_handler = \
        AntEnsemble(self,
                    exp_name,
                    self.data_factory,
                    self.dump_dir,
                    self.api_token,
                    self.task,
                    self.dataset)
      # 准备ensemble控制
      ensemble_handler.start()
      yield ensemble_handler

      # 关闭recorder
      ensemble_handler.context.recorder.close()

      # 日志更新
      mlogger.update()

      # 环境回收
      ensemble_handler.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Demo(self, exp_name, exp_param, html_template=None,  **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)

    if 'config' in kwargs:
      params = self.params._params
      params.update({
        'demo': kwargs['config']
      })
      self.params = params

    try:
      demo = AntDemo(self,
                     exp_name,
                     self.dump_dir,
                     self.api_token,
                     self.task,
                     html_template=html_template,
                     time_stamp=timestamp())
      # demo
      demo.start()
      yield demo

      if kwargs.get('clear', True):
        # 日志更新
        mlogger.update()

        # 环境回收
        demo.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Browser(self, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)

    if 'config' in kwargs:
      params = self.params._params
      params.update({
        'browser': kwargs['config']
      })
      self.params = params

    try:
      browser = AntBrowser(self,
                           exp_name,
                           self.api_token,
                           self.data_factory,
                           self.dataset,
                           self.dump_dir)
      # 浏览控制
      browser.start()
      yield browser

      if kwargs.get('clear', True):
        # 日志更新
        mlogger.update()

        # 环境回收
        browser.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  @contextmanager
  def Activelearning(self, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    self._prepare_context(exp_name, exp_param, **kwargs)

    if 'config' in kwargs:
      params = self.params._params
      params.update({
        'activelearning': kwargs['config']
      })
      self.params = params

    try:
      activelearning = AntActiveLearningV2(
        self,
        exp_name,
        self.data_factory,
        self.dump_dir,
        self.api_token,
        self.task)
      # 浏览控制
      activelearning.start()
      yield activelearning

      if kwargs.get('clear', True):
        # 日志更新
        mlogger.update()

        # 环境回收
        activelearning.context.wait_until_clear()
    except:
      traceback.print_exc()
      raise sys.exc_info()[0]

  def __getattr__(self, __name: str):
    assert(__name in ['ensemble', 'browser', 'demo', 'activelearning'])
    if __name in self.server_map:
      return self.server_map[__name]

    cc = self
    class Server(object):
      def __init__(self, name) -> None:
        self.name = name
        self.handler = None
        cc.server_map[self.name] = self

      def start(self, exp_name, exp_param={}, **kwargs):
        if self.name == 'ensemble':
          cc._prepare_context(exp_name, exp_param, **kwargs)

          if 'config' in kwargs:
            params = cc.params._params
            params.update({
              'ensemble': kwargs['config']
            })
            cc.params = params

          self.handler = \
            AntEnsemble(cc,
                        exp_name,
                        cc.data_factory,
                        cc.dump_dir,
                        cc.api_token,
                        cc.task,
                        cc.dataset)
          # 准备ensemble控制
          self.handler.start()
        elif self.name == 'browser':
          # 1.step 配置基本信息
          cc._prepare_context(exp_name, exp_param, **kwargs)

          if 'config' in kwargs:
            if 'size' not in kwargs['config']:
              kwargs['config'].update({'size': 0})
            params = cc.params._params
            params.update({
              'browser': kwargs['config']
            })
            cc.params = params

          self.handler = AntBrowser(cc,
                                exp_name,
                                cc.data_factory,
                                cc.dump_dir,
                                cc.api_token,   
                                cc.task,                             
                                cc.dataset)
          # 浏览控制
          self.handler.start(**kwargs)
        elif self.name == 'demo':
          # 1.step 配置基本信息
          cc._prepare_context(exp_name, exp_param, **kwargs)

          if 'config' in kwargs:
            params = cc.params._params
            params.update({
              'demo': kwargs['config']
            })
            cc.params = params

            self.handler = AntDemo(cc,
                          exp_name,
                          cc.dump_dir,
                          cc.api_token,
                          cc.task,
                          html_template=kwargs.get('html_template', ''),
                          time_stamp=timestamp())
            # demo
            self.handler.start()          
        elif self.name == 'activelearning':
          # 1.step 配置基本信息
          cc._prepare_context(exp_name, exp_param, **kwargs)

          if 'config' in kwargs:
            params = cc.params._params
            params.update({
              'activelearning': kwargs['config']
            })
            cc.params = params

            self.handler = AntActiveLearningV2(
              cc,
              exp_name,
              cc.data_factory,
              cc.dump_dir,
              cc.api_token,
              cc.task)
            self.handler.start(**kwargs)          

      def exit(self):
        if self.handler is not None:
          # 关闭recorder
          self.handler.context.recorder.close()
          # 环境回收
          self.handler.context.wait_until_clear()
      
      def __getattr__(self, __name: str):
        return getattr(self.handler, __name)

    return Server(__name)


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import grad
from ...builder import DISTILL_LOSSES


'''
LwM is originally an incremental learning method with 
classification/distillation/attention distillation losses.
Here, LwM is only defined as the Grad-CAM based attention distillation.
'''

@DISTILL_LOSSES.register_module()
class LwM(nn.Module):
	'''
	Learning without Memorizing
	https://arxiv.org/pdf/1811.08051.pdf
	'''
	def __init__(self):
		super(LwM, self).__init__()

	def forward(self, out_s, fm_s, out_t, fm_t, target):
		target_out_t = torch.gather(out_t, 1, target.view(-1, 1))
		grad_fm_t    = grad(outputs=target_out_t, inputs=fm_t,
							grad_outputs=torch.ones_like(target_out_t),
							create_graph=True, retain_graph=True, only_inputs=True)[0]
		weights_t = F.adaptive_avg_pool2d(grad_fm_t, 1)
		cam_t = torch.sum(torch.mul(weights_t, grad_fm_t), dim=1, keepdim=True)
		cam_t = F.relu(cam_t)
		cam_t = cam_t.view(cam_t.size(0), -1)
		norm_cam_t = F.normalize(cam_t, p=2, dim=1)

		target_out_s = torch.gather(out_s, 1, target.view(-1, 1))
		grad_fm_s    = grad(outputs=target_out_s, inputs=fm_s,
							grad_outputs=torch.ones_like(target_out_s),
							create_graph=True, retain_graph=True, only_inputs=True)[0]
		weights_s = F.adaptive_avg_pool2d(grad_fm_s, 1)
		cam_s = torch.sum(torch.mul(weights_s, grad_fm_s), dim=1, keepdim=True)
		cam_s = F.relu(cam_s)
		cam_s = cam_s.view(cam_s.size(0), -1)
		norm_cam_s = F.normalize(cam_s, p=2, dim=1)

		loss = F.l1_loss(norm_cam_s, norm_cam_t.detach())

		return loss


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class Logits(nn.Module):
	'''
	Do Deep Nets Really Need to be Deep?
	http://papers.nips.cc/paper/5484-do-deep-nets-really-need-to-be-deep.pdf
	'''
	def __init__(self):
		super(Logits, self).__init__()

	def forward(self, out_s, out_t):
		loss = F.mse_loss(out_s, out_t)

		return loss


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES

@DISTILL_LOSSES.register_module()
class HCL(nn.Module):
    def __init__(self):
        super(HCL, self).__init__()

    def forward(self, fstudent, fteacher):
        loss_all = 0.0
        for fs, ft in zip(fstudent, fteacher):
            n,c,h,w = fs.shape
            loss = F.mse_loss(fs, ft, reduction='mean')
            cnt = 1.0
            tot = 1.0
            for l in [4,2,1]:
                if l >=h:
                    continue
                tmpfs = F.adaptive_avg_pool2d(fs, (l,l))
                tmpft = F.adaptive_avg_pool2d(ft, (l,l))
                cnt /= 2.0
                loss += F.mse_loss(tmpfs, tmpft, reduction='mean') * cnt
                tot += cnt
            loss = loss / tot
            loss_all = loss_all + loss
        return loss_all



# encoding=utf-8
# @Time    : 22-3-7
# @File    : help.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import os
import re
import git
import logging
import shutil
from antgo import config
import json
from pprint import pprint
import time
import yaml
from antgo.framework.helper.utils import Config
from antgo.ant.client import *


def prepare_project_environment(project_git, project_branch, project_commit):
    git_folder = project_git.split('/')[-1].split('.')[0]
    if os.path.exists(git_folder):
        logging.warn('Project code existed.')
        return
        
    # 无代码环境，处理流程
    logging.info(f"Clone project code from {project_git}")
    if project_branch is not None:
        if project_commit is not None:
            os.system(f'git clone {project_git} -b {project_branch}; git reset --soft {project_commit}')
        else:
            os.system(f'git clone {project_git}; -b {project_branch};')
    else:
        os.system(f'git clone {project_git}')
    
    os.chdir(f'./{git_folder}')


def check_project_environment(args):
    assert(args.project != '')
    if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
        logging.error(f'Project {args.project} not exist.')
        return False
    
    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
        project_info = json.load(fp)

    # 检查是否在项目目录下
    is_in_project_folder = False
    try:
        rep = git.Repo('./')
        is_in_project_folder = True
        cur_git_address = ''
        for url_address in rep.remote().urls:
            cur_git_address = url_address
            break
        if cur_git_address.split('/')[-1] != project_info['git'].split('/')[-1]:
            logging.error('You are in an error project folder.')
            return False

        if rep.is_dirty() or len(rep.untracked_files) > 0:
            logging.error(f'Some uncomitted/untracked files in repo.')
            return False        
    except:
        is_in_project_folder = False

    project_git = project_info['git']
    if not is_in_project_folder:
        # 无代码环境，处理流程
        git_folder = project_git.split('/')[-1].split('.')[0]
        if not os.path.exists(git_folder): 
            if args.branch is not None:
                if args.commit is not None:
                    os.system(f'git clone {project_git} -b {args.branch}; git reset --soft {args.commit}')
                else:
                    os.system(f'git clone {project_git}; -b {args.branch};')
            else:
                os.system(f'git clone {project_git}')

        os.chdir(f'./{git_folder}')

    return True


def generate_project_exp_example(template_project_folder, target_folder, exp_name):
    # step1: 拷贝cifar10分类模型代码
    if exp_name is None:
        exp_name = 'cifar10'
    if not os.path.exists(os.path.join(target_folder, exp_name)):
        # 复制cifar10分类样例代码
        if exp_name is None or exp_name == '':
            exp_name = 'cifar10'
        shutil.copytree(os.path.join(template_project_folder, 'cifar10'), os.path.join(target_folder, exp_name))

        if exp_name != 'cifar10':
            # 修改main.py的帮助信息
            with open( os.path.join(target_folder, exp_name, 'main.py'), 'r') as fp:
                content = fp.read()
                content = content.replace('cifar10', exp_name)
            
            with open( os.path.join(target_folder, exp_name, 'main.py'), 'w') as fp:
                fp.write(content)

            # 修改标准配置文件 (lsp, pascal_voc)
            os.remove(os.path.join(target_folder, exp_name, 'configs', 'config.py'))
            if os.path.exists(os.path.join(template_project_folder, f'{exp_name}_config.py')):
                shutil.copy(os.path.join(template_project_folder, f'{exp_name}_config.py'), os.path.join(target_folder, exp_name, 'configs', 'config.py'))
            else:      
               shutil.copy(os.path.join(template_project_folder, 'config.py'), os.path.join(target_folder, exp_name, 'configs'))
    else:
        logging.warn('MVP code has existed in current path.')
        return

    # step2: 拷贝shell等说明代码
    if not os.path.exists(os.path.join(target_folder, 'system.py')):
        shutil.copy(os.path.join(template_project_folder, 'system.py'), target_folder)
    if not os.path.exists(os.path.join(target_folder, 'install.sh')):
        shutil.copy(os.path.join(template_project_folder, 'install.sh'), target_folder)
    if not os.path.exists(os.path.join(target_folder, 'launch.sh')):
        shutil.copy(os.path.join(template_project_folder, 'launch.sh'), target_folder)
    if not os.path.exists(os.path.join(target_folder, 'requirements.txt')):
        shutil.copy(os.path.join(template_project_folder, 'requirements.txt'), target_folder)
    if not os.path.exists(os.path.join(target_folder, 'README.md')):
        # 修改readme信息
        with open( os.path.join(template_project_folder, 'README.md'), 'r') as fp:
            content = fp.read()
            content = content.replace('cifar10', exp_name)

        with open( os.path.join(target_folder, 'README.md'), 'w') as fp:
            fp.write(content)

    if not os.path.exists(os.path.join(target_folder, '.gitignore')):
        shutil.copy(os.path.join(template_project_folder, '.gitignore'), target_folder)


def project_add_action(action_name, args):
    if args.project == '':
        logging.error('Must set --project')
        return
    
    if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
        logging.error(f'Project {args.project} not existed.')
        return        

    get_lock().acquire()
    project_info = {}
    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
        project_info = json.load(fp) 

    if action_name == 'expert':
        if args.exp not in project_info['exp']:
            logging.error(f'Exp {args.exp} not in project {args.project}.')
            get_lock().release()
            return

        is_finding_exp = False
        for exp_info in project_info['exp'][args.exp]:
            if args.id == exp_info['id']:
                is_finding_exp = True
                break

        if not is_finding_exp:
            logging.error(f'Exp {args.exp} ID {args.id} not in project exp list.')
            get_lock().release()
            return
        
        exp_with_id = f'{args.exp}.{args.id}'
        if exp_with_id in project_info['expert']:
            logging.error(f'Exp {args.exp} ID {args.id} has existed in project {args.project} expert list')
            get_lock().release()
            return

        project_info['expert'].append(exp_with_id)
        pprint(project_info['expert'])
    elif action_name == 'product':
        if args.exp not in project_info['exp']:
            logging.error(f'Exp {args.exp} not in project exp list.')
            get_lock().release()
            return 
        project_info['product'] = args.exp

        # 因为要触发后台，任务调度服务，提前保存项目信息
        with open(os.path.join(config.AntConfig.task_factory, f'{args.project}.json'), 'w') as fp:
            json.dump(project_info, fp)  
        
        pprint(project_info['product'])
        get_lock().release()
        ######### trigger:start #########
        # 重新训练产品模型（监督训练->半监督训练->蒸馏训练）
        get_client().trigger(f'{args.project}.product')
        ######### trigger:end   #########        
        return
    elif action_name == 'baseline':
        if args.exp not in project_info['exp']:
            logging.error(f'Exp {args.exp} not in project {args.project}.')
            get_lock().release()
            return

        is_finding_exp = False
        for exp_info in project_info['exp'][args.exp]:
            if args.id == exp_info['id']:
                is_finding_exp = True
                break

        if not is_finding_exp:
            logging.error(f'Exp {args.exp} ID {args.id} not in project exp list.')
            get_lock().release()
            return

        exp_with_id = f'{args.exp}.{args.id}'
        project_info['baseline'] = exp_with_id
        pprint(project_info['baseline'])
    elif action_name == 'train/label':
        if args.address is None:
            logging.error('Must set --address')
            get_lock().release()
            return
        
        if not file_client_exists(f'{args.address}-index') or not file_client_exists(f'{args.address}-tfrecord'):
            logging.error(f'Data {args.address} dont exist')
            get_lock().release()
            return
        
        project_info['dataset']['train']['label'].append(
            {
                "tag": args.tags,
                "num": args.num,
                "status": True,
                'address': args.address
            }
        )

        # 因为要触发后台，任务调度服务，提前保存项目信息
        with open(os.path.join(config.AntConfig.task_factory, f'{args.project}.json'), 'w') as fp:
            json.dump(project_info, fp)  

        pprint(project_info['dataset']['train']['label'])
        get_lock().release()
        ######### trigger:start #########
        # 基于更新后的有标签数据，基于当前最佳产品模型，重新监督训练/蒸馏训练
        get_client().trigger(f'{args.project}.train/label')
        ######### trigger:end   #########
        return
    elif action_name == 'train/unlabel':
        if args.address is None:
            logging.error('Must set --address')
            get_lock().release()
            return
        
        if not file_client_exists(f'{args.address}-index') or not file_client_exists(f'{args.address}-tfrecord'):
            logging.error(f'Data {args.address} dont exist')
            get_lock().release()
            return

        project_info['dataset']['train']['unlabel'].append(
            {
                "tag": args.tags,
                "num": args.num,
                "status": True,
                'address': args.address
            }
        )

        # 因为要触发后台，任务调度服务，提前保存项目信息
        with open(os.path.join(config.AntConfig.task_factory, f'{args.project}.json'), 'w') as fp:
            json.dump(project_info, fp)  
        
        pprint(project_info['dataset']['train']['unlabel'])
        get_lock().release()
        ######### trigger:start #########
        # 基于更新后的无标签数据，基于当前最佳产品模型，重新半监督模型训练
        get_client().trigger(f'{args.project}.train/unlabel')
        ######### trigger:end   #########
        return
    elif action_name == 'test':
        if args.address is None:
            logging.error('Must set --address')
            get_lock().release()
            return
                
        project_info['dataset']['test'] = {
            "tag": args.tags,
            "num": args.num,
            "status": True,
            'address': args.address
        }
        pprint(project_info['dataset']['test'])
    elif action_name == 'val':
        if args.address is None:
            logging.error('Must set --address')
            get_lock().release()
            return
        
        project_info['dataset']['val'] = {
            "tag": args.tags,
            "num": args.num,
            "status": True,
            'address': args.address
        }
        project_info['dataset']['val']
    else:
        logging.error(f'Command {action_name} dont support.')

    # 在本地存储项目信息
    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
        json.dump(project_info, fp)
    get_lock().release()


def project_del_action(action_name, args):
    if args.project == '':
        logging.error('Must set --project')
        return

    if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
        logging.error(f'Project {args.project} not existed.')
        return

    get_lock().acquire()
    project_info = {}
    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
        project_info = json.load(fp) 

    if action_name == 'expert':
        if args.exp not in project_info['exp']:
            logging.error(f'Exp {args.exp} not in project {args.project}.')
            get_lock().release()
            return
        
        exp_with_id = f'{args.exp}.{args.id}'
        if exp_with_id not in project_info['expert']:
            logging.error(f'{exp_with_id} not exist in project {args.project}')
            get_lock().release()
            return

        project_info['expert'].remove(exp_with_id)
        logging.info(f'Del expert {exp_with_id} of project {args.project}')
    elif action_name == 'train/label':
        after_list = []
        for info in project_info['dataset']['train']['label']:
            if info['tag'] == args.tags:
                continue
            after_list.append(info)
        
        project_info['dataset']['train']['label'] = after_list
        logging.info(f'Del train/label with tags {args.tags} of project {args.project}')
    elif action_name == 'train/unlabel':
        after_list = []
        for info in project_info['dataset']['train']['unlabel']:
            if info['tag'] == args.tags:
                continue
            after_list.append(info)
        
        project_info['dataset']['train']['unlabel'] = after_list
        logging.info(f'Del train/unlabel with tags {args.tags} of project {args.project}')
    elif action_name == 'test':
        project_info['dataset']['test'] = {}
        logging.info(f'Del test dataset of project {args.project}')
    elif action_name == 'val':
        project_info['dataset']['val'] = {}
        logging.info(f'Del val dataset of project {args.project}')
    else:
        logging.error(f'Command {action_name} dont support.')

    # 在本地存储项目信息
    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
        json.dump(project_info, fp)    

    get_lock().release()


def show_action(action_name, args):
    if action_name == 'best':
        if args.project == '':
            logging.error('Must set --project')
            return

        if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
            logging.error(f'Project {args.project} not existed.')
            return   
        
        project_info = {}
        with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
            project_info = json.load(fp) 
                    
        pprint(project_info['best'])
    elif action_name == 'project':
        if args.project == '':
            all_projects = []
            for project_file in os.listdir(config.AntConfig.task_factory):
                if not project_file.endswith('.json'):
                    continue
                all_projects.append(project_file.replace('.json', ''))

            pprint(all_projects)
        else:
            if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
                logging.error(f'Project {args.project} not existed.')
                return

            project_info = {}
            with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
                project_info = json.load(fp) 

            pprint(project_info)
    elif action_name == "exp":
        if args.project == '':
            logging.error('Must set --project')
            return

        if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
            logging.error(f'Project {args.project} not existed.')
            return   

        project_info = {}
        with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
            project_info = json.load(fp) 

        pprint(project_info['exp'])


def get_action(action_name, args):
    logging.error('Now dont support')


def update_project_config(sub_action_name, args):
    if args.project == '':
        logging.error('Must set --project')
        return

    if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
        logging.error(f'Project {args.project} not existed.')
        return

    get_lock().acquire()
    project_info = {}
    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
        project_info = json.load(fp) 

    if sub_action_name == 'project/semi':
        # 更新半监督方案配置
        # args.name 来检查是否是内置算法
        if args.name == '' and args.config == '':
            # 清空配置
            project_info['tool']['semi']['method'] = ''
            project_info['tool']['semi']['config'] = dict()

            with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
                json.dump(project_info, fp)
            logging.info(f'Success remove {sub_action_name} config')  
            get_lock().release()
            return    
        if not os.path.exists(args.config):
            if args.name not in ['dense', 'detmpl', 'mpl']:
                logging.error(f'Missing config file, {sub_action_name} couldnt config')
                get_lock().release()
                return
            else:
                args.config = os.path.join(os.path.dirname(__file__), 'framework', 'helper', 'configs', sub_action_name.split('/')[-1], f'{args.name}_config.py')

        if not os.path.exists(args.config):
            logging.error(f'Config file {args.config} not exist')
            get_lock().release()
            return

        # 读取配置
        assert(args.config.endswith('.py'))
        cfg = Config.fromfile(args.config)

        # 设置半监督算法的名字        
        project_info['tool']['semi']['method'] = args.name
        # 设置半监督的配置
        project_info['tool']['semi']['config'].update(
            cfg._cfg_dict.to_dict()
        )

        # print info
        pprint(project_info['tool']['semi'])
    elif sub_action_name == 'project/distillation':
        # args.name 来检查是否是内置算法
        if args.name == '' and args.config == '':
            # 清空配置
            project_info['tool']['distillation']['method'] = ''
            project_info['tool']['distillation']['config'] = dict()

            with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
                json.dump(project_info, fp)
            logging.info(f'Success remove {sub_action_name} config')    
            get_lock().release()          
            return    

        if not os.path.exists(args.config):
            if args.name not in ['reviewkd']:
                logging.error(f'Missing config file, {sub_action_name} couldnt config')
                get_lock().release()
                return
            else:
                args.config = os.path.join(os.path.dirname(__file__), 'framework', 'helper', 'configs', sub_action_name.split('/')[-1], f'{args.name}_config.py')

        if not os.path.exists(args.config):
            logging.error(f'Config file {args.config} not exist')
            get_lock().release()
            return

        # 读取配置
        assert(args.config.endswith('.py'))
        cfg = Config.fromfile(args.config)

        # 设置蒸馏算法的名字        
        project_info['tool']['distillation']['method'] = args.name
        # 设置蒸馏的配置
        project_info['tool']['distillation']['config'].update(
            cfg._cfg_dict.to_dict()
        )

        # print info
        pprint(project_info['tool']['distillation'])
    elif sub_action_name == 'project/activelearning':
        # 更新主动学习方案配置
        if args.name == '' and args.config == '':
            # 清空配置
            project_info['tool']['activelearning']['method'] = ''
            project_info['tool']['activelearning']['config'] = dict()

            with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
                json.dump(project_info, fp)
            logging.info(f'Success remove {sub_action_name} config') 
            get_lock().release()             
            return    

        if not os.path.exists(args.config):
            if args.name not in ['dense', 'detmpl', 'mpl', 'ac']:
                logging.error(f'Missing config file, {sub_action_name} couldnt config')
                get_lock().release()
                return
            else:
                args.config = os.path.join(os.path.dirname(__file__), 'framework', 'helper', 'configs', sub_action_name.split('/')[-1], f'{args.name}_config.py')

        if not os.path.exists(args.config):
            logging.error(f'Config file {args.config} not exist')
            get_lock().release()
            return

        # 读取配置
        assert(args.config.endswith('.py'))
        cfg = Config.fromfile(args.config)

        # 设置主动学习算法的名字        
        project_info['tool']['activelearning']['method'] = args.name
        # 设置主动学习的配置
        project_info['tool']['activelearning']['config'].update(
            cfg._cfg_dict.to_dict()
        )
        
        # print info
        pprint(project_info['tool']['activelearning'])
    elif sub_action_name == 'project/label':
        assert(args.config.endswith('.py'))
        cfg = Config.fromfile(args.config)

        cfg.type = cfg.type.upper()
        if cfg.type not in ['', 'CLASS', 'RECT','POINT','POLYGON', 'SKELETON']:
            logging.error(f"Dont support label type {cfg.type}")
            get_lock().release()
            return

        if not isinstance(cfg.category, list):
            logging.error(f"Label category must be list type")
            get_lock().release()
            return

        if cfg.type == 'SKELETON':
            # 对于关键点标注，需要添加meta/skeleton信息
            if cfg.get('meta', None):
                logging.error('Must set meta/skeleton info for SKELETON label')
                get_lock().release()
                return

            if len(cfg.meta.get('skeleton', [])) == 0:
                logging.error('Must set meta/skeleton info for SKELETON label')
                get_lock().release()
                return

        project_info['tool']['label']['category'] = cfg.category    # 标注类别 ['clsname', 'clsname', ...]
        project_info['tool']['label']['type'] = cfg.type            # 标注类型
        if cfg.get('meta', None):
            project_info['tool']['label']['meta']['skeleton'] = cfg.meta.get('skeleton', [])

        # print info
        pprint(project_info['tool']['label'])
    elif sub_action_name == 'project/submitter':
        if args.ssh:
            project_info['submitter']['method'] = 'ssh'
        elif args.local:
            project_info['submitter']['method'] = 'local'
        else:
            # 检查本地是否存在定制化提交脚本
            submit_config_file = os.path.join(os.environ['HOME'], '.config', 'antgo', 'submit-config.yaml')    
            if not os.path.exists(submit_config_file):
                logging.error('No custom submit script config')
                get_lock().release()
                return

            with open(submit_config_file, encoding='utf-8', mode='r') as fp:
                config_content = yaml.safe_load(fp)
            script_folder = config_content['folder']
            script_file = config_content['script']
            if not os.path.exists(os.path.join(script_folder, script_file)):
                logging.error('Custom submit scrip launch file not exist.')
                get_lock().release()
                return

            project_info['submitter']['method'] = 'custom'

        if args.gpu >= 0:
            project_info['submitter']['gpu_num'] = args.gpu
        if args.cpu >= 0:
            project_info['submitter']['cpu_num'] = args.cpu
        if args.memory >= 0:
            project_info['submitter']['memory'] = args.memory

        # print info
        pprint(project_info['submitter'])
    else:
        logging.error(f"Dont support {sub_action_name}")
        get_lock().release()
        return

    with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
        json.dump(project_info, fp)

    logging.info(f'Success update {sub_action_name} config')    
    get_lock().release()



from .file_client import (BaseStorageBackend, AliBackend, FileClient, file_client_get, file_client_put, file_client_ls,file_client_mkdir,file_client_rm,file_client_exists )
from .handlers import BaseFileHandler, JsonHandler, PickleHandler, YamlHandler
from .io import dump, load, register_handler
from .parse import dict_from_file, list_from_file

__all__ = [
    'BaseStorageBackend', 'AliBackend', 'FileClient', 'load', 'dump', 'register_handler',
    'BaseFileHandler', 'JsonHandler', 'PickleHandler', 'YamlHandler',
    'list_from_file', 'dict_from_file', 'file_client_get', 'file_client_put', 'file_client_ls',
    'file_client_mkdir', 'file_client_rm', 'file_client_exists'
]



# -*- coding: UTF-8 -*-
# @Time : 2018/8/24
# @File : coco2017.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import copy
import itertools
import json
import time
import numpy as np
from collections import defaultdict
import sys
from urllib.request import urlretrieve
from antgo.dataflow.dataset.dataset import *
from antgo.utils import mask as maskUtils
from antgo.utils.fs import maybe_here_match_format
from antgo.framework.helper.fileio.file_client import *


__all__ = ['COCO2017']
class CocoAPI():
  def __init__(self, annotation_file=None):
    """
    Constructor of Microsoft COCO helper class for reading and visualizing annotations.
    :param annotation_file (str): location of annotation file
    :param image_folder (str): location to the folder that hosts images.
    :return:
    """
    # load dataset
    self.dataset, self.anns, self.cats, self.imgs = dict(), dict(), dict(), dict()
    self.imgToAnns, self.catToImgs = defaultdict(list), defaultdict(list)
    if not annotation_file == None:
      print('loading annotations into memory...')
      tic = time.time()
      dataset = json.load(open(annotation_file, 'r'))
      assert type(dataset) == dict, 'annotation file format {} not supported'.format(type(dataset))
      print('Done (t={:0.2f}s)'.format(time.time() - tic))
      self.dataset = dataset
      self.createIndex()

  def createIndex(self):
    # create index
    print('creating index...')
    anns, cats, imgs = {}, {}, {}
    imgToAnns, catToImgs = defaultdict(list), defaultdict(list)
    if 'annotations' in self.dataset:
      for ann in self.dataset['annotations']:
        if 'segments_info' not in ann:
          imgToAnns[ann['image_id']].append(ann)
          anns[ann['id']] = ann
        else:
          imgToAnns[ann['image_id']].append(ann['segments_info'])
          for mm in ann['segments_info']:
            anns[mm['id']] = mm

    if 'images' in self.dataset:
      for img in self.dataset['images']:
        imgs[img['id']] = img

    if 'categories' in self.dataset:
      for cat in self.dataset['categories']:
        cats[cat['id']] = cat

    if 'annotations' in self.dataset and 'categories' in self.dataset:
      for ann in self.dataset['annotations']:
        if 'segments_info' not in ann:
          catToImgs[ann['category_id']].append(ann['image_id'])
        else:
          for mm in ann['segments_info']:
            catToImgs[mm['category_id']] = ann['image_id']

    print('index created!')

    # create class members
    self.anns = anns
    self.imgToAnns = imgToAnns
    self.catToImgs = catToImgs
    self.imgs = imgs
    self.cats = cats

  def info(self):
    """
    Print information about the annotation file.
    :return:
    """
    for key, value in self.dataset['info'].items():
      print('{}: {}'.format(key, value))

  def getAnnIds(self, imgIds=[], catIds=[], areaRng=[], iscrowd=None):
    """
    Get ann ids that satisfy given filter conditions. default skips that filter
    :param imgIds  (int array)     : get anns for given imgs
           catIds  (int array)     : get anns for given cats
           areaRng (float array)   : get anns for given area range (e.g. [0 inf])
           iscrowd (boolean)       : get anns for given crowd label (False or True)
    :return: ids (int array)       : integer array of ann ids
    """
    imgIds = imgIds if type(imgIds) == list else [imgIds]
    catIds = catIds if type(catIds) == list else [catIds]

    if len(imgIds) == len(catIds) == len(areaRng) == 0:
      anns = self.dataset['annotations']
    else:
      if not len(imgIds) == 0:
        lists = [self.imgToAnns[imgId] for imgId in imgIds if imgId in self.imgToAnns]
        anns = list(itertools.chain.from_iterable(lists))
      else:
        anns = self.dataset['annotations']
      anns = anns if len(catIds) == 0 else [ann for ann in anns if ann['category_id'] in catIds]
      anns = anns if len(areaRng) == 0 else [ann for ann in anns if
                                             ann['area'] > areaRng[0] and ann['area'] < areaRng[1]]
    if not iscrowd == None:
      ids = [ann['id'] for ann in anns if ann['iscrowd'] == iscrowd]
    else:
      ids = [ann['id'] for ann in anns]
    return ids

  def getCatIds(self, catNms=[], supNms=[], catIds=[]):
    """
    filtering parameters. default skips that filter.
    :param catNms (str array)  : get cats for given cat names
    :param supNms (str array)  : get cats for given supercategory names
    :param catIds (int array)  : get cats for given cat ids
    :return: ids (int array)   : integer array of cat ids
    """
    catNms = catNms if type(catNms) == list else [catNms]
    supNms = supNms if type(supNms) == list else [supNms]
    catIds = catIds if type(catIds) == list else [catIds]

    if len(catNms) == len(supNms) == len(catIds) == 0:
      cats = self.dataset['categories']
    else:
      cats = self.dataset['categories']
      cats = cats if len(catNms) == 0 else [cat for cat in cats if cat['name'] in catNms]
      cats = cats if len(supNms) == 0 else [cat for cat in cats if cat['supercategory'] in supNms]
      cats = cats if len(catIds) == 0 else [cat for cat in cats if cat['id'] in catIds]
    ids = [cat['id'] for cat in cats]
    return ids

  def getImgIds(self, imgIds=[], catIds=[]):
    '''
    Get img ids that satisfy given filter conditions.
    :param imgIds (int array) : get imgs for given ids
    :param catIds (int array) : get imgs with all given cats
    :return: ids (int array)  : integer array of img ids
    '''
    imgIds = imgIds if type(imgIds) == list else [imgIds]
    catIds = catIds if type(catIds) == list else [catIds]

    if len(imgIds) == len(catIds) == 0:
      ids = self.imgs.keys()
    else:
      ids = set(imgIds)
      for i, catId in enumerate(catIds):
        if i == 0 and len(ids) == 0:
          ids = set(self.catToImgs[catId])
        else:
          ids |= set(self.catToImgs[catId])
    return list(ids)

  def loadAnns(self, ids=[]):
    """
    Load anns with the specified ids.
    :param ids (int array)       : integer ids specifying anns
    :return: anns (object array) : loaded ann objects
    """
    if type(ids) == list:
      return [self.anns[id] for id in ids]
    elif type(ids) == int:
      return [self.anns[ids]]

  def loadCats(self, ids=[]):
    """
    Load cats with the specified ids.
    :param ids (int array)       : integer ids specifying cats
    :return: cats (object array) : loaded cat objects
    """
    if type(ids) == list:
      return [self.cats[id] for id in ids]
    elif type(ids) == int:
      return [self.cats[ids]]

  def loadImgs(self, ids=[]):
    """
    Load anns with the specified ids.
    :param ids (int array)       : integer ids specifying img
    :return: imgs (object array) : loaded img objects
    """
    if type(ids) == list:
      return [self.imgs[id] for id in ids]
    elif type(ids) == int:
      return [self.imgs[ids]]

  def loadRes(self, resFile):
    """
    Load result file and return a result api object.
    :param   resFile (str)     : file name of result file
    :return: res (obj)         : result api object
    """
    res = CocoAPI()
    res.dataset['images'] = [img for img in self.dataset['images']]

    print('Loading and preparing results...')
    tic = time.time()
    if type(resFile) == str or type(resFile) == unicode:
      anns = json.load(open(resFile))
    elif type(resFile) == np.ndarray:
      anns = self.loadNumpyAnnotations(resFile)
    else:
      anns = resFile
    assert type(anns) == list, 'results in not an array of objects'
    annsImgIds = [ann['image_id'] for ann in anns]
    assert set(annsImgIds) == (set(annsImgIds) & set(self.getImgIds())), \
      'Results do not correspond to current coco set'
    if 'caption' in anns[0]:
      imgIds = set([img['id'] for img in res.dataset['images']]) & set([ann['image_id'] for ann in anns])
      res.dataset['images'] = [img for img in res.dataset['images'] if img['id'] in imgIds]
      for id, ann in enumerate(anns):
        ann['id'] = id + 1
    elif 'bbox' in anns[0] and not anns[0]['bbox'] == []:
      res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
      for id, ann in enumerate(anns):
        bb = ann['bbox']
        x1, x2, y1, y2 = [bb[0], bb[0] + bb[2], bb[1], bb[1] + bb[3]]
        if not 'segmentation' in ann:
          ann['segmentation'] = [[x1, y1, x1, y2, x2, y2, x2, y1]]
        ann['area'] = bb[2] * bb[3]
        ann['id'] = id + 1
        ann['iscrowd'] = 0
    elif 'segmentation' in anns[0]:
      res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
      for id, ann in enumerate(anns):
        # now only support compressed RLE format as segmentation results
        ann['area'] = maskUtils.area(ann['segmentation'])
        if not 'bbox' in ann:
          ann['bbox'] = maskUtils.toBbox(ann['segmentation'])
        ann['id'] = id + 1
        ann['iscrowd'] = 0
    elif 'keypoints' in anns[0]:
      res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
      for id, ann in enumerate(anns):
        s = ann['keypoints']
        x = s[0::3]
        y = s[1::3]
        x0, x1, y0, y1 = np.min(x), np.max(x), np.min(y), np.max(y)
        ann['area'] = (x1 - x0) * (y1 - y0)
        ann['id'] = id + 1
        ann['bbox'] = [x0, y0, x1 - x0, y1 - y0]
    print('DONE (t={:0.2f}s)'.format(time.time() - tic))

    res.dataset['annotations'] = anns
    res.createIndex()
    return res

  def download(self, tarDir=None, imgIds=[]):
    '''
    Download COCO images from mscoco.org server.
    :param tarDir (str): COCO results directory name
           imgIds (list): images to be downloaded
    :return:
    '''
    if tarDir is None:
      print('Please specify target directory')
      return -1
    if len(imgIds) == 0:
      imgs = self.imgs.values()
    else:
      imgs = self.loadImgs(imgIds)
    N = len(imgs)
    if not os.path.exists(tarDir):
      os.makedirs(tarDir)
    for i, img in enumerate(imgs):
      tic = time.time()
      fname = os.path.join(tarDir, img['file_name'])
      if not os.path.exists(fname):
        urlretrieve(img['coco_url'], fname)
      print('downloaded {}/{} images (t={:0.1f}s)'.format(i, N, time.time() - tic))

  def loadNumpyAnnotations(self, data):
    """
    Convert result data from a numpy array [Nx7] where each row contains {imageID,x1,y1,w,h,score,class}
    :param  data (numpy.ndarray)
    :return: annotations (python nested list)
    """
    print('Converting ndarray to lists...')
    assert (type(data) == np.ndarray)
    print(data.shape)
    assert (data.shape[1] == 7)
    N = data.shape[0]
    ann = []
    for i in range(N):
      if i % 1000000 == 0:
        print('{}/{}'.format(i, N))
      ann += [{
        'image_id': int(data[i, 0]),
        'bbox': [data[i, 1], data[i, 2], data[i, 3], data[i, 4]],
        'score': data[i, 5],
        'category_id': int(data[i, 6]),
      }]
    return ann

  def annToRLE(self, ann):
    """
    Convert annotation which can be polygons, uncompressed RLE to RLE.
    :return: binary mask (numpy 2D array)
    """
    t = self.imgs[ann['image_id']]
    h, w = t['height'], t['width']
    segm = ann['segmentation']
    if type(segm) == list:
      # polygon -- a single object might consist of multiple parts
      # we merge all parts into one mask rle code
      rles = maskUtils.frPyObjects(segm, h, w)
      rle = maskUtils.merge(rles)
    elif type(segm['counts']) == list:
      # uncompressed RLE
      rle = maskUtils.frPyObjects(segm, h, w)
    else:
      # rle
      rle = ann['segmentation']
    return rle

  def annToMask(self, ann):
    """
    Convert annotation which can be polygons, uncompressed RLE, or RLE to binary mask.
    :return: binary mask (numpy 2D array)
    """
    rle = self.annToRLE(ann)
    m = maskUtils.decode(rle)
    return m


class COCO2017(Dataset):
  METAINFO = {
        'classes':
        ('person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train',
         'truck', 'boat', 'traffic light', 'fire hydrant', 'stop sign',
         'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep',
         'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella',
         'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard',
         'sports ball', 'kite', 'baseball bat', 'baseball glove', 'skateboard',
         'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup', 'fork',
         'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange',
         'broccoli', 'carrot', 'hot dog', 'pizza', 'donut', 'cake', 'chair',
         'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
         'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave',
         'oven', 'toaster', 'sink', 'refrigerator', 'book', 'clock', 'vase',
         'scissors', 'teddy bear', 'hair drier', 'toothbrush'),
        # palette is a list of color tuples, which is used for visualization.
        'palette':
        [(220, 20, 60), (119, 11, 32), (0, 0, 142), (0, 0, 230), (106, 0, 228),
         (0, 60, 100), (0, 80, 100), (0, 0, 70), (0, 0, 192), (250, 170, 30),
         (100, 170, 30), (220, 220, 0), (175, 116, 175), (250, 0, 30),
         (165, 42, 42), (255, 77, 255), (0, 226, 252), (182, 182, 255),
         (0, 82, 0), (120, 166, 157), (110, 76, 0), (174, 57, 255),
         (199, 100, 0), (72, 0, 118), (255, 179, 240), (0, 125, 92),
         (209, 0, 151), (188, 208, 182), (0, 220, 176), (255, 99, 164),
         (92, 0, 73), (133, 129, 255), (78, 180, 255), (0, 228, 0),
         (174, 255, 243), (45, 89, 255), (134, 134, 103), (145, 148, 174),
         (255, 208, 186), (197, 226, 255), (171, 134, 1), (109, 63, 54),
         (207, 138, 255), (151, 0, 95), (9, 80, 61), (84, 105, 51),
         (74, 65, 105), (166, 196, 102), (208, 195, 210), (255, 109, 65),
         (0, 143, 149), (179, 0, 194), (209, 99, 106), (5, 121, 0),
         (227, 255, 205), (147, 186, 208), (153, 69, 1), (3, 95, 161),
         (163, 255, 0), (119, 0, 170), (0, 182, 199), (0, 165, 120),
         (183, 130, 88), (95, 32, 0), (130, 114, 135), (110, 129, 133),
         (166, 74, 118), (219, 142, 185), (79, 210, 114), (178, 90, 62),
         (65, 70, 15), (127, 167, 115), (59, 105, 106), (142, 108, 45),
         (196, 172, 0), (95, 54, 80), (128, 76, 255), (201, 57, 1),
         (246, 0, 122), (191, 162, 208)]
  }

  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(COCO2017, self).__init__(train_or_test, dir, ext_params)
    self.year = "2017"
    self.dir = dir
    self.train_or_test = train_or_test
    self.data_type = None
    self.is_support_random_split = False
    self.task_type = getattr(self, 'task_type', None)
    self.task_type_subset = getattr(self, 'task_type_subset', 'stuff')
    self.task_test = getattr(self, 'task_test', None)

    assert(self.train_or_test in ['train', 'val', 'test'])
    assert (self.task_type in ['SEGMENTATION', 'OBJECT-DETECTION', 'INSTANCE-SEGMENTATION', 'LANDMARK'])

    if not os.path.exists(os.path.join(self.dir , 'annotations')):
      if os.environ.get('LOCAL_RANK', 0) == 0:
        if not os.path.exists(os.path.join(self.dir, 'COCO')):
          # 数据集不存在，需要重新下载，并创建标记
          ali = AliBackend()
          ali.download('ali:///dataset/coco/COCO.tar', self.dir)
          assert(os.path.exists(os.path.join(self.dir, 'COCO.tar')))
          # 解压
          os.system(f'cd {self.dir} && tar -xf COCO.tar')
          os.system('touch DATASET_IS_READY')
        else:
          # 数据集存在，创建标记
          if not os.path.exists('DATASET_IS_READY'):
            os.system('touch DATASET_IS_READY')
      else:
        while True:
          # 等待直到存在指定文件
          if os.path.exists('DATASET_IS_READY'):
            break
          time.sleep(5)

      # 修改数据目录
      self.dir = os.path.join(self.dir, 'COCO')

    data_type = None
    if self.train_or_test == "train":
      data_type = "train" + self.year
    elif self.train_or_test == "val":
      data_type = "val" + self.year
    elif self.train_or_test == "test":
      data_type = "test" + self.year
    self.data_type = data_type

    if self.train_or_test in ['train', 'val']:
      if self.task_type in ["OBJECT-DETECTION", "INSTANCE-SEGMENTATION"]:
        # annotation file
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        # parse (for object detector)
        # self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
        self.cat_ids = self.coco_api.getCatIds(catNms=list(self.METAINFO['classes']))
        self.img_ids = self.coco_api.getImgIds(catIds=self.cat_ids)
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
      elif self.task_type == "LANDMARK":
        # parse (for object detector)
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        # parse (for person task)
        self.cat_ids = self.coco_api.getCatIds(catNms=['person'])
        self.img_ids = self.coco_api.getImgIds(catIds=self.cat_ids)

        # annotation file
        kps_ann_file = self.config_ann_file(data_type, self.dir, 'PersonKeypoint')
        self.coco_kps_api = CocoAPI(kps_ann_file)
      elif self.task_type == "IMAGE_CAPTION":
        # parse (for object detector)
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
        self.cat_ids = self.coco_api.getCatIds(self.cats)
        self.img_ids = self.coco_api.getImgIds(self.cat_ids)

        # parse (for caption)
        caps_ann_file = self.config_ann_file(data_type, self.dir, 'Caption')
        self.coco_caps_api = CocoAPI(caps_ann_file)
      elif self.task_type == "SEGMENTATION":
        # Panoptic Segmentation
        assert(self.task_type_subset in ['stuff', 'panoptic'])
        if self.task_type_subset == 'stuff':
          ann_file = os.path.join(self.dir, 'annotations', 'stuff_%s2017.json'%self.train_or_test)
          self.coco_api = CocoAPI(ann_file)

          self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
          self.cat_ids = self.coco_api.getCatIds(self.cats)
          self.img_ids = self.coco_api.getImgIds(self.cat_ids)
        else:
          ann_file = os.path.join(self.dir, 'annotations', 'panoptic_%s2017.json'%self.train_or_test)
          self.coco_api = CocoAPI(ann_file)
          self.img_ids = self.coco_api.getImgIds()
      else:
        # annotation file
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        # parse (for object detector)
        self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
        self.cat_ids = self.coco_api.getCatIds(self.cats)
        self.img_ids = self.coco_api.getImgIds(self.cat_ids)
    else:
      ann_file = ''

      if self.task_test == 'dev':
        ann_file = os.path.join(self.dir, 'annotations', 'image_info_test-dev2017.json')
      else:
        ann_file = os.path.join(self.dir, 'annotations', 'image_info_test2017.json')

      self.coco_api = CocoAPI(ann_file)

      # parse (for object detector)
      self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
      self.cat_ids = self.coco_api.getCatIds(self.cats)
      self.img_ids = self.coco_api.getImgIds(self.cat_ids)

  def config_ann_file(self, data_type, data_dir, annotation_type):
    maybe_data_dir = maybe_here_match_format(data_dir, 'annotations')
    assert maybe_data_dir is not None

    ann_file = None
    if annotation_type == "Instance":
      ann_file = '%s/annotations/instances_%s.json' % (maybe_data_dir, data_type)
    elif annotation_type == "PersonKeypoint":
      ann_file = '%s/annotations/person_keypoints_%s.json' % (maybe_data_dir, data_type)
    elif annotation_type == "Caption":
      ann_file = '%s/annotations/captions_%s.json' % (maybe_data_dir, data_type)

    return ann_file

  def at(self, id):
    img_obj = self.coco_api.loadImgs(self.img_ids[id])[0]
    if self.train_or_test == 'test':
      # 对于测试集没有标签
      img = imread(os.path.join(self.dir, '%s2017' % self.train_or_test, img_obj['file_name']))
      return (img, {})

    if self.task_type == 'OBJECT-DETECTION' or \
            self.task_type=='INSTANCE-SEGMENTATION':
      annotation_ids = self.coco_api.getAnnIds(imgIds=img_obj['id'])
      annotation = self.coco_api.loadAnns(annotation_ids)
      img_annotation = {}

      boxes = []
      category_id = []
      for ix, obj in enumerate(annotation):
        x, y, w, h = obj['bbox']        
        inter_w = max(0, min(x + w, img_obj['width']) - max(x, 0))
        inter_h = max(0, min(y + h, img_obj['height']) - max(y, 0))
        if inter_w * inter_h == 0:
                continue   
        if obj['area'] <= 0 or w < 1 or h < 1:
            continue
        if obj['category_id'] not in self.cat_ids:
            continue
        # 目标框
        boxes.append([x, y, x + w, y + h])
        # 目标类别
        category_id.append(self.cat2label[obj['category_id']])
        # 忽略目标分割

      img_annotation['bboxes'] = np.array(boxes)
      img_annotation['labels'] = np.array(category_id)
      img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
      img_annotation['image_meta'] = {
        'image_shape': (img.shape[0], img.shape[1])
      }

      return (img, img_annotation)
    elif self.task_type == 'SEGMENTATION':
      # stuff, Panoptic
      img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
      annotation_ids = self.coco_api.getAnnIds(imgIds=img_obj['id'])
      annotation = self.coco_api.loadAnns(annotation_ids)
      category_id = np.zeros((len(annotation)), dtype=np.int32)
      segmentation = []
      for ix, obj in enumerate(annotation):
        category_id[ix] = self.cat2label(obj['category_id'])
        segmentation.append(self.coco_api.annToMask(obj))

      segmentation_map = np.zeros((img.shape[0], img.shape[1]), dtype=np.int32)
      for ix, obj_seg in enumerate(segmentation):
        obj_id = category_id[ix]
        segmentation_map[np.where(obj_seg == 1)] = obj_id

      img_annotation = {
        'segments': segmentation_map,
        'image_meta': {
          'image_shape': (img.shape[0], img.shape[1])
        }
      }
      return (img, img_annotation)
    elif self.task_type == 'LANDMARK':
      img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
      ann_ids = self.coco_kps_api.getAnnIds(imgIds=img_obj['id'])
      anns = self.coco_kps_api.loadAnns(ann_ids)

      boxes = []
      labels = []
      joints2d = []
      joints_vis = []

      for person_ann in anns:
        keypionts = person_ann['keypoints']        
        person_keypoints_xy = np.zeros((len(keypionts)//3, 2))
        person_keypoints_xy[:, 0] = keypionts[0::3]
        person_keypoints_xy[:, 1] = keypionts[1::3]

        keypoints_visible = np.array(keypionts[2::3])
        position_visible = np.where(keypoints_visible>0)
        if position_visible[0].size == 0:
          continue        
        keypoints_visible[position_visible] = 1
        person_bbox = [
          np.min(person_keypoints_xy[position_visible, 0]), 
          np.min(person_keypoints_xy[position_visible, 1]), 
          np.max(person_keypoints_xy[position_visible, 0]), 
          np.max(person_keypoints_xy[position_visible, 1])
        ]
        if person_bbox[2] - person_bbox[0] <= 5 or person_bbox[3] - person_bbox[1] <= 5:
          continue

        joints2d.append(person_keypoints_xy)
        joints_vis.append(keypoints_visible)
        boxes.append(person_bbox)    
        labels.append(0)  # person

      # 空，跳过
      if len(joints2d) == 0:
        return (None, None)

      img_annotation = {
        'joints2d': np.stack(joints2d, 0),
        'joints_vis': np.stack(joints_vis, 0),
        'bboxes': np.array(boxes),
        'labels': np.array(labels),
        'image_meta': {
          'image_shape': (img.shape[0], img.shape[1])
        }
      }
      return (img, img_annotation)

    return (None, None)

  @property
  def size(self):
    return len(self.img_ids)

  def split(self, train_validation_ratio=0.0, is_stratified_sampling=True):
    assert(self.train_or_test == 'train')
    validation_coco = COCO2017('val', self.dir, self.ext_params)
    return self, validation_coco

# coco2017 = COCO2017('train', '/root/workspace/dataset/COCO', ext_params={'task_type': 'OBJECT-DETECTION'})
# label_max = 0
# for i in range(coco2017.size):
#   data = coco2017.sample(i)
#   # if data['labels'].size > 0:
#   #   label_max = max(label_max, np.max(data['labels']))
#   #   print(f'label_max {label_max}')
#   print(i)

#   image = data['image']
#   for bi in range(len(data['bboxes'])):
#     x0,y0,x1,y1 = data['bboxes'][bi]
#     cls_label = data['labels'][bi]
#     x0=(int)(x0)
#     y0=(int)(y0)
#     x1=(int)(x1)
#     y1=(int)(y1)
    
#     color_v = coco2017.METAINFO['palette'][cls_label]
#     image = cv2.rectangle(image, (x0,y0),(x1,y1), color_v, 4)
#   cv2.imwrite("./crop_show.png", image)


import argparse
import os.path as osp
import warnings
from functools import partial
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.runner.checkpoint import load_checkpoint
from antgo.framework.helper.tools.util import *
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.utils import Config
import numpy as np
import onnx
import os
import torch

def parse_args():
    parser = argparse.ArgumentParser(
        description='Convert ANTGO models to ONNX')
    parser.add_argument('--config', type=str, default='/root/paddlejob/workspace/env_run/portrait/InterHand26M/main/test_config.py', help='test config file path')
    parser.add_argument('--ext-module', type=str, default='ext_module.py', help='introduce ext module py file')
    parser.add_argument('--checkpoint', type=str, default='/root/paddlejob/workspace/env_run/portrait/InterHand26M/main/main/latest.pth', help='checkpoint file')
    parser.add_argument('--output-file', type=str, default='tmp.onnx')
    parser.add_argument('--opset-version', type=int, default=11)
    parser.add_argument(
        '--shape',
        type=int,
        nargs='+',
        default=[1, 3, 224, 224],
        help='input image size')
    parser.add_argument(
        '--input-name',
        type=str,
        default='input'
    )
    parser.add_argument(
        '--output-names',
        type=str,
        default='output'
    )
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()
    cfg = Config.fromfile(args.config)

    if args.ext_module != '':
        # 导入ext module
        print(f'load ext model from {args.ext_module}')
        load_extmodule(args.ext_module)

    print('build model')
    cfg.model.train_cfg = None
    model = \
        build_model(cfg.model, test_cfg=cfg.get('test_cfg', None))
    print(f"load chekcpoint from {args.checkpoint}")
    load_checkpoint(model, args.checkpoint, map_location='cpu')

    input_name = args.input_name
    output_names = args.output_names.split(',')
    input_data = np.ones(args.shape, dtype=np.float32)
    
    print('export onnx')
    torch.onnx.export(
        model,
        input_data,
        args.output_file,
        input_names=[input_name],
        output_names=output_names,
        export_params=True,
        keep_initializers_as_inputs=True,
        do_constant_folding=True,
        verbose=True,
        opset_version=args.opset_version)
    




# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 21:56
# @File    : computer_vision.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import cv2
import numpy as np
import os
import json


class ComputerVisionMixin:
  """
  Mixin for computer vision problems.
  """
  def image_imshow(self, title='image'):  # pragma: no cover
    for im in self:
      cv2.imshow(title, im)
      cv2.waitKey(0)

  @classmethod
  def read_camera(cls, device_id=0, limit=-1):  # pragma: no cover
    """
    read images from a camera.
    """
    cnt = limit

    def inner():
      nonlocal cnt
      cap = cv2.VideoCapture(device_id)
      while cnt != 0:
        retval, im = cap.read()
        if retval:
          yield im
          cnt -= 1

    return cls(inner())

  # pylint: disable=redefined-builtin
  @classmethod
  def read_video(cls, path, format='rgb24'):
    """
    Load video as a datacollection.

    Args:
        path:
            The path to the target video.
        format:
            The format of the images loaded from video.
    """
    def inner():
      cap = cv2.VideoCapture(path)
      while cap.isOpened():
        ret, frame = cap.read()
        if not ret:
          break
        yield frame

    return cls(inner())

  def map_group(self, output_path, rows=10, cols=10):
    count = 0
    big_image = None
    init_width = -1
    init_height = -1
    shard = 0

    for x in self:
      row_i = count // cols
      col_i = (int)(count - row_i*cols)

      if len(x.shape) == 2:
        x = np.stack([x,x,x], -1)
      x = x[:,:,:3]
      height, width = x.shape[:2]
      big_width = (int)(cols * width)
      big_height = (int)(rows * height)
      if init_width < 0:
        init_width = width
      if init_height < 0:
        init_height = height

      if big_width != (int)(cols * init_width):
        big_width = (int)(cols * init_width)
      if big_height != (int)(rows * init_height):
        big_height = (int)(rows * init_height)

      if big_image is None:
        big_image = np.zeros((big_height, big_width, 3), dtype=np.uint8) 
      elif (row_i+1)*height > big_image.shape[0]:
        # 保存
        if not os.path.exists(output_path):
          os.makedirs(output_path)
        cv2.imwrite(os.path.join(output_path, f'{shard}.png'), big_image)
        shard += 1

        big_image = np.zeros((big_height, big_width, 3), dtype=np.uint8) 
        count = 0
        row_i = count // cols
        col_i = (int)(count - row_i*cols)
      
      # 填充当前图片
      big_image[row_i*height: row_i*height+x.shape[0], col_i*width:col_i*width+x.shape[1]] = x
      count += 1

  def to_video(self,
               output_path,
               width,
               height,
               rate=30,
               codec=None,
               format=None,
               template=None,
               audio_src=None):
    """
    Encode a video with audio if provided.

    Args:
        output_path:
            The path of the output video.
        codec:
            The codec to encode and decode the video.
        rate:
            The rate of the video.
        width:
            The width of the video.
        height:
            The height of the video.
        format:
            The format of the video frame image.
        template:
            The template video stream of the ouput video stream.
        audio_src:
            The audio to encode with the video.
    """
    assert(width is not None and height is not None)
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')  # 用于mp4格式的生成的参数
    out = cv2.VideoWriter(output_path, fourcc, rate, (width, height))  # 创建一个写入视频对象

    for array in self:
      h,w = array.shape[:2]
      if h != height or w !=width:
        array = cv2.resize(array, (width, height))
      out.write(array)

    out.release()

  def to_json(self, path):
    total = []
    for data in self:
      assert(isinstance(data, dict))
      total.append(data)
      
    with open(path, 'w') as fp:
      json.dump(total, fp)


# Copyright (c) OpenMMLab. All rights reserved.
import torch

from ..builder import BBOX_ASSIGNERS
from ..iou_calculators import build_iou_calculator
from ..transforms import bbox_xyxy_to_cxcywh
from .assign_result import AssignResult
from .base_assigner import BaseAssigner


@BBOX_ASSIGNERS.register_module()
class UniformAssigner(BaseAssigner):
    """Uniform Matching between the anchors and gt boxes, which can achieve
    balance in positive anchors, and gt_bboxes_ignore was not considered for
    now.

    Args:
        pos_ignore_thr (float): the threshold to ignore positive anchors
        neg_ignore_thr (float): the threshold to ignore negative anchors
        match_times(int): Number of positive anchors for each gt box.
           Default 4.
        iou_calculator (dict): iou_calculator config
    """

    def __init__(self,
                 pos_ignore_thr,
                 neg_ignore_thr,
                 match_times=4,
                 iou_calculator=dict(type='BboxOverlaps2D')):
        self.match_times = match_times
        self.pos_ignore_thr = pos_ignore_thr
        self.neg_ignore_thr = neg_ignore_thr
        self.iou_calculator = build_iou_calculator(iou_calculator)

    def assign(self,
               bbox_pred,
               anchor,
               gt_bboxes,
               gt_bboxes_ignore=None,
               gt_labels=None):
        num_gts, num_bboxes = gt_bboxes.size(0), bbox_pred.size(0)

        # 1. assign -1 by default
        assigned_gt_inds = bbox_pred.new_full((num_bboxes, ),
                                              0,
                                              dtype=torch.long)
        assigned_labels = bbox_pred.new_full((num_bboxes, ),
                                             -1,
                                             dtype=torch.long)
        if num_gts == 0 or num_bboxes == 0:
            # No ground truth or boxes, return empty assignment
            if num_gts == 0:
                # No ground truth, assign all to background
                assigned_gt_inds[:] = 0
            assign_result = AssignResult(
                num_gts, assigned_gt_inds, None, labels=assigned_labels)
            assign_result.set_extra_property(
                'pos_idx', bbox_pred.new_empty(0, dtype=torch.bool))
            assign_result.set_extra_property('pos_predicted_boxes',
                                             bbox_pred.new_empty((0, 4)))
            assign_result.set_extra_property('target_boxes',
                                             bbox_pred.new_empty((0, 4)))
            return assign_result

        # 2. Compute the L1 cost between boxes
        # Note that we use anchors and predict boxes both
        cost_bbox = torch.cdist(
            bbox_xyxy_to_cxcywh(bbox_pred),
            bbox_xyxy_to_cxcywh(gt_bboxes),
            p=1)
        cost_bbox_anchors = torch.cdist(
            bbox_xyxy_to_cxcywh(anchor), bbox_xyxy_to_cxcywh(gt_bboxes), p=1)

        # We found that topk function has different results in cpu and
        # cuda mode. In order to ensure consistency with the source code,
        # we also use cpu mode.
        # TODO: Check whether the performance of cpu and cuda are the same.
        C = cost_bbox.cpu()
        C1 = cost_bbox_anchors.cpu()

        # self.match_times x n
        index = torch.topk(
            C,  # c=b,n,x c[i]=n,x
            k=self.match_times,
            dim=0,
            largest=False)[1]

        # self.match_times x n
        index1 = torch.topk(C1, k=self.match_times, dim=0, largest=False)[1]
        # (self.match_times*2) x n
        indexes = torch.cat((index, index1),
                            dim=1).reshape(-1).to(bbox_pred.device)

        pred_overlaps = self.iou_calculator(bbox_pred, gt_bboxes)
        anchor_overlaps = self.iou_calculator(anchor, gt_bboxes)
        pred_max_overlaps, _ = pred_overlaps.max(dim=1)
        anchor_max_overlaps, _ = anchor_overlaps.max(dim=0)

        # 3. Compute the ignore indexes use gt_bboxes and predict boxes
        ignore_idx = pred_max_overlaps > self.neg_ignore_thr
        assigned_gt_inds[ignore_idx] = -1

        # 4. Compute the ignore indexes of positive sample use anchors
        # and predict boxes
        pos_gt_index = torch.arange(
            0, C1.size(1),
            device=bbox_pred.device).repeat(self.match_times * 2)
        pos_ious = anchor_overlaps[indexes, pos_gt_index]
        pos_ignore_idx = pos_ious < self.pos_ignore_thr

        pos_gt_index_with_ignore = pos_gt_index + 1
        pos_gt_index_with_ignore[pos_ignore_idx] = -1
        assigned_gt_inds[indexes] = pos_gt_index_with_ignore

        if gt_labels is not None:
            assigned_labels = assigned_gt_inds.new_full((num_bboxes, ), -1)
            pos_inds = torch.nonzero(
                assigned_gt_inds > 0, as_tuple=False).squeeze()
            if pos_inds.numel() > 0:
                assigned_labels[pos_inds] = gt_labels[
                    assigned_gt_inds[pos_inds] - 1]
        else:
            assigned_labels = None

        assign_result = AssignResult(
            num_gts,
            assigned_gt_inds,
            anchor_max_overlaps,
            labels=assigned_labels)
        assign_result.set_extra_property('pos_idx', ~pos_ignore_idx)
        assign_result.set_extra_property('pos_predicted_boxes',
                                         bbox_pred[indexes])
        assign_result.set_extra_property('target_boxes',
                                         gt_bboxes[pos_gt_index])
        return assign_result



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.cnn.bricks.wrappers import NewEmptyTensorOp, obsolete_torch_version

if torch.__version__ == 'parrots':
    TORCH_VERSION = torch.__version__
else:
    # torch.__version__ could be 1.3.1+cu92, we only need the first two
    # for comparison
    TORCH_VERSION = tuple(int(x) for x in torch.__version__.split('.')[:2])


def adaptive_avg_pool2d(input, output_size):
    """Handle empty batch dimension to adaptive_avg_pool2d.

    Args:
        input (tensor): 4D tensor.
        output_size (int, tuple[int,int]): the target output size.
    """
    if input.numel() == 0 and obsolete_torch_version(TORCH_VERSION, (1, 9)):
        if isinstance(output_size, int):
            output_size = [output_size, output_size]
        output_size = [*input.shape[:2], *output_size]
        empty = NewEmptyTensorOp.apply(input, output_size)
        return empty
    else:
        return F.adaptive_avg_pool2d(input, output_size)


class AdaptiveAvgPool2d(nn.AdaptiveAvgPool2d):
    """Handle empty batch dimension to AdaptiveAvgPool2d."""

    def forward(self, x):
        # PyTorch 1.9 does not support empty tensor inference yet
        if x.numel() == 0 and obsolete_torch_version(TORCH_VERSION, (1, 9)):
            output_size = self.output_size
            if isinstance(output_size, int):
                output_size = [output_size, output_size]
            else:
                output_size = [
                    v if v is not None else d
                    for v, d in zip(output_size,
                                    x.size()[-2:])
                ]
            output_size = [*x.shape[:2], *output_size]
            empty = NewEmptyTensorOp.apply(x, output_size)
            return empty

        return super().forward(x)



import torch
import torch.nn as nn
from torch.nn import functional as F
import math
from antgo.framework.helper.runner import BaseModule
from antgo.framework.helper.cnn import ConvModule, DepthwiseSeparableConvModule
from antgo.framework.helper.cnn.bricks.waterfall import PWWaterFall, ConvWaterFall
from antgo.framework.helper.models.builder import HEADS
from antgo.framework.helper.models.pose3d.head.layer import make_linear_layers


@HEADS.register_module()
class TwoHand3DPoseLatent(BaseModule):
    def __init__(self, joint_num=21, output_root_hm_shape=24, **kwargs):
        super(TwoHand3DPoseLatent, self).__init__()
        self.joint_num = joint_num  # single hand
        self.output_root_hm_shape = output_root_hm_shape
        # 左手热图
        self.joint_left_enc = nn.Sequential(
            ConvWaterFall(64, 32, 64),
            ConvWaterFall(64, 32, 64)
        )
        self.joint_left_final_layer = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 21, kernel_size=1, bias=True)
        )     

        # 左手latent depth-maps
        self.joint_left_latent_depth = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 21, kernel_size=1, bias=True)         
        )

        # 右手热图+偏移
        self.joint_right_enc = nn.Sequential(
            ConvWaterFall(64, 32, 64),
            ConvWaterFall(64, 32, 64)
        )
        self.joint_right_final_layer = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 21, kernel_size=1, bias=True)
        )
        # 右手latent depth-maps
        self.joint_right_latent_depth = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 21, kernel_size=1, bias=True)            
        )        

        self.root_fc = make_linear_layers([1024,1024,self.output_root_hm_shape],relu_final=False) # 2048->512
        self.hand_fc = make_linear_layers([1024,1024,2],relu_final=False)     

    def soft_argmax_1d(self, heatmap1d):
        heatmap1d = F.softmax(heatmap1d, 1)
        if heatmap1d.device.type == 'cpu':
            accu = heatmap1d * torch.arange(self.output_root_hm_shape).float()[None,:]
        else:
            accu = heatmap1d * torch.arange(self.output_root_hm_shape).float().cuda()[None,:]
            
        coord = accu.sum(dim=1)
        return coord

    def forward(self, img_feat):
        # left and right hand share
        left_x32, right_x32, x4 = img_feat

        # 左手heatmap + latent
        left_up32 = self.joint_left_enc(left_x32)
        joint_left_h = self.joint_left_final_layer(left_up32)
        joint_left_latent_depth = self.joint_left_latent_depth(left_up32)

        # 右手heatmap + latent
        right_up32 = self.joint_right_enc(right_x32)
        joint_right_h = self.joint_right_final_layer(right_up32)
        joint_right_latent_depth = self.joint_right_latent_depth(right_up32)
        
        # 左右手相对深度
        img_feat = x4
        img_feat_gap = F.avg_pool2d(img_feat, (img_feat.shape[2],img_feat.shape[3])).view(-1, 1024)
        root_heatmap1d = self.root_fc(img_feat_gap)
        root_depth = self.soft_argmax_1d(root_heatmap1d).view(-1,1)
        
        # 左右手可见性
        hand_type = torch.sigmoid(self.hand_fc(img_feat_gap))   

        left_depth_and_sigma = None
        right_depth_and_sigma = None
        left_uv_off_and_sigma = None
        right_uv_off_and_sigma = None
        return joint_left_h, joint_left_latent_depth, \
                joint_right_h, joint_right_latent_depth, \
                left_depth_and_sigma, \
                right_depth_and_sigma, \
                left_uv_off_and_sigma, \
                right_uv_off_and_sigma, \
                root_depth, hand_type